import dolfin
import numpy as np
import pulse
import pytest
import simcardems


@pytest.mark.xfail
@pytest.mark.parametrize(
    "pre_stretch, traction, spring, fix_right_plane, expected_u",
    [
        (None, None, None, True, (0, 0, 0)),
        (None, None, None, False, (0, 0, 0)),
        (None, -10.0, None, True, (0, 0, 0)),
        (None, -10.0, None, False, (0.21933731, -0.04719835, -0.04719835)),
        (0.1, None, None, True, (0.1, -0.02326871, -0.02326871)),
        (0.1, None, None, False, (0.1, -0.02326871, -0.02326871)),
        (None, None, 100, False, (0, 0, 0)),
    ],
)
def test_boundary_conditions(
    pre_stretch,
    traction,
    spring,
    fix_right_plane,
    expected_u,
):
    mesh = dolfin.UnitCubeMesh(2, 2, 2)

    bcs, marker_functions = simcardems.mechanics_model.setup_diriclet_bc(
        mesh,
        pre_stretch=pre_stretch,
        traction=traction,
        spring=spring,
        fix_right_plane=fix_right_plane,
    )
    microstructure = simcardems.mechanics_model.setup_microstructure(mesh)
    geometry = pulse.Geometry(
        mesh=mesh,
        microstructure=microstructure,
        marker_functions=marker_functions,
    )
    material = pulse.NeoHookean()
    problem = pulse.MechanicsProblem(
        geometry,
        material,
        bcs,
        solver_parameters={"linear_solver": "mumps"},
    )
    problem.solve()

    if traction is not None:
        pulse.iterate.iterate(problem, problem.bcs.neumann[0].traction, traction)

    U, P = problem.state.split(deepcopy=True)

    print(U(1.0, 0.5, 0.5))
    assert np.isclose(U(1.0, 0.5, 0.5), expected_u, atol=1e-5).all()


@pytest.mark.xfail
@pytest.mark.parametrize(
    "pre_stretch, traction, spring, fix_right_plane, expected_u",
    [
        (None, None, None, True, (0, 0, 0)),
        (None, None, None, False, (-0.15656733, 0.04443344, 0.04443344)),
        (None, -10.0, None, True, (0.0, 0, 0)),
        (None, -10.0, None, False, (0.0, 0.0, 0.0)),
        (0.1, None, None, True, (0.1, -0.02326871, -0.02326871)),
        (0.1, None, None, False, (0.1, -0.02326871, -0.02326871)),
        (None, None, 100, True, (0, 0, 0)),
        (None, None, 100, False, (-0.0615345, 0.00249483, 0.00249483)),
    ],
)
def test_boundary_conditions_activation(
    pre_stretch,
    traction,
    spring,
    fix_right_plane,
    expected_u,
):
    mesh = dolfin.UnitCubeMesh(2, 2, 2)

    bcs, marker_functions = simcardems.mechanics_model.setup_diriclet_bc(
        mesh,
        pre_stretch=pre_stretch,
        traction=traction,
        spring=spring,
        fix_right_plane=fix_right_plane,
    )
    microstructure = simcardems.mechanics_model.setup_microstructure(mesh)
    geometry = pulse.Geometry(
        mesh=mesh,
        microstructure=microstructure,
        marker_functions=marker_functions,
    )
    Ta = dolfin.Constant(0.0)
    material = pulse.NeoHookean(activation=Ta, active_model="active_stress")
    problem = pulse.MechanicsProblem(
        geometry,
        material,
        bcs,
        solver_parameters={"linear_solver": "mumps"},
    )
    problem.solve()

    if traction is not None:
        pulse.iterate.iterate(problem, problem.bcs.neumann[0].traction, traction)

    pulse.iterate.iterate(problem, Ta, 10.0)

    U, P = problem.state.split(deepcopy=True)
    print(U(1.0, 0.5, 0.5))
    assert np.isclose(U(1.0, 0.5, 0.5), expected_u, rtol=1e-5).all()


if __name__ == "__main__":
    test_boundary_conditions_activation(None, None, 100, True, None)



from __future__ import annotations

import json
from pathlib import Path
from typing import Dict
from typing import Optional
from typing import Type
from typing import TYPE_CHECKING

import cbcbeat
import dolfin
import pulse
import ufl

from . import geometry
from . import utils
from .config import Config

if TYPE_CHECKING:
    from .models.em_model import BaseEMCoupling
    from .models.cell_model import BaseCellModel

logger = utils.getLogger(__name__)


def setup_cell_model(
    cls,
    coupling: BaseEMCoupling,
    cell_params=None,
    cell_inits=None,
    cell_init_file=None,
    drug_factors_file=None,
    popu_factors_file=None,
    disease_state=Config.disease_state,
):
    cell_params = handle_cell_params(
        cell_params=cell_params,
        disease_state=disease_state,
        drug_factors_file=drug_factors_file,
        popu_factors_file=popu_factors_file,
        CellModel=cls,
    )

    cell_inits = handle_cell_inits(
        cell_inits=cell_inits,
        cell_init_file=cell_init_file,
        CellModel=cls,
    )

    return cls(
        init_conditions=cell_inits,
        params=cell_params,
        coupling=coupling,
    )


def setup_solver(
    dt,
    cellmodel: cbcbeat.CardiacCellModel,
    coupling: BaseEMCoupling,
    scheme=Config.ep_ode_scheme,
    theta=Config.ep_theta,
    preconditioner=Config.ep_preconditioner,
    PCL=Config.PCL,
) -> cbcbeat.SplittingSolver:
    # Set-up cardiac model
    ps = setup_splitting_solver_parameters(
        theta=theta,
        preconditioner=preconditioner,
        dt=dt,
        scheme=scheme,
    )
    ep_heart = setup_model(
        cellmodel,
        coupling.geometry.ep_mesh,
        PCL=PCL,
        microstructure=coupling.geometry.microstructure_ep,
        stimulus_domain=coupling.geometry.stimulus_domain,
    )
    solver = cbcbeat.SplittingSolver(ep_heart, params=ps)
    # Extract the solution fields and set the initial conditions
    (vs_, vs, vur) = solver.solution_fields()
    vs_.assign(cellmodel.initial_conditions())

    coupling.register_ep_model(solver)
    coupling.print_ep_info()
    return solver


def default_conductivities():
    # Conductivities as defined by page 4339 of Niederer benchmark
    return dict(
        sigma_il=0.17,  # mS / mm
        sigma_it=0.019,  # mS / mm
        sigma_el=0.62,  # mS / mm
        sigma_et=0.24,  # mS / mm
    )


def define_conductivity_tensor(
    microstructure: pulse.Microstructure,
    chi: float = 140.0,
    C_m: float = 0.01,
):
    fiber = microstructure.f0
    sheet = microstructure.s0
    cross_sheet = microstructure.n0

    A = dolfin.as_matrix(
        [
            [fiber[0], sheet[0], cross_sheet[0]],
            [fiber[1], sheet[1], cross_sheet[1]],
            [fiber[2], sheet[2], cross_sheet[2]],
        ],
    )

    # FIXME: Make is possible to change this from the outside
    conductivities = default_conductivities()

    # Compute monodomain approximation by taking harmonic mean in each
    # direction of intracellular and extracellular part
    def harmonic_mean(a, b):
        return a * b / (a + b)

    sigma_l = harmonic_mean(conductivities["sigma_il"], conductivities["sigma_el"])
    sigma_t = harmonic_mean(conductivities["sigma_it"], conductivities["sigma_et"])

    # Scale conductivities by 1/(C_m * chi)
    s_l = sigma_l / (C_m * chi)  # mm^2 / ms
    s_t = sigma_t / (C_m * chi)  # mm^2 / ms

    # Define conductivity tensor
    M_star = ufl.diag(dolfin.as_vector([s_l, s_t, s_t]))
    M = A * M_star * A.T

    return M


def setup_model(
    cellmodel: cbcbeat.CardiacCellModel,
    mesh: dolfin.Mesh,
    microstructure: pulse.Microstructure,
    stimulus_domain: geometry.StimulusDomain,
    PCL: int = 1000,
    chi: float = 140.0,
    C_m: float = 0.01,
    duration: float = 2.0,
    A: float = 50_000.0,
) -> cbcbeat.CardiacModel:
    """Set-up cardiac model based on benchmark parameters


    Parameters
    ----------
    cellmodel : cbcbeat.CardiacCellModel
        _description_
    mesh : dolfin.Mesh
        _description_
    microstructure : pulse.Microstructure
        _description_
    PCL : int, optional
        _description_, by default 1000
    chi : float, optional
        Surface to volume ratio in mm^{-1}, by default 140.0
    C_m : float, optional
        Membrane capacitance in  mu F / mm^2, by default 0.01
    duration: float, optional
        Stimulation duration in milliseconds, by default 2.0
    A: float, optional
        FIXME: Some value in mu A/cm^3, by default 50_000.0

    Returns
    -------
    cbcbeat.CardiacModel
        The EP model
    """
    # Define time
    time = dolfin.Constant(0.0)

    # Define conductivity tensor
    M = define_conductivity_tensor(chi=chi, C_m=C_m, microstructure=microstructure)

    # Define stimulation (NB: region of interest carried by the mesh
    # and assumptions in cbcbeat)
    cm2mm = 10.0
    factor = 1.0 / (chi * C_m)  # NB: cbcbeat convention
    amplitude = factor * A * (1.0 / cm2mm) ** 3  # mV/ms

    s = "((std::fmod(time,PCL) >= start) & (std::fmod(time,PCL) <= duration + start)) ? amplitude : 0.0"

    I_s = dolfin.Expression(
        s,
        time=time,
        start=0.0,
        duration=duration,
        amplitude=amplitude,
        PCL=PCL,
        degree=0,
    )
    # Store input parameters in cardiac model
    stimulus = cbcbeat.Markerwise(
        (I_s,),
        (stimulus_domain.marker,),
        stimulus_domain.domain,
    )

    petsc_options = [
        ["ksp_type", "cg"],
        ["pc_type", "gamg"],
        ["pc_gamg_verbose", "10"],
        ["pc_gamg_square_graph", "0"],
        ["pc_gamg_coarse_eq_limit", "3000"],
        ["mg_coarse_pc_type", "redundant"],
        ["mg_coarse_sub_pc_type", "lu"],
        ["mg_levels_ksp_type", "richardson"],
        ["mg_levels_ksp_max_it", "3"],
        ["mg_levels_pc_type", "sor"],
    ]
    for opt in petsc_options:
        dolfin.PETScOptions.set(*opt)

    heart = cbcbeat.CardiacModel(
        domain=mesh,
        time=time,
        M_i=M,
        M_e=None,
        cell_models=cellmodel,
        stimulus=stimulus,
        applied_current=None,
    )

    return heart


def setup_splitting_solver_parameters(
    dt,
    theta=0.5,
    preconditioner="sor",
    scheme="GRL1",
):
    ps = cbcbeat.SplittingSolver.default_parameters()
    ps["pde_solver"] = "monodomain"
    ps["MonodomainSolver"]["linear_solver_type"] = "iterative"
    ps["MonodomainSolver"]["theta"] = theta
    ps["MonodomainSolver"]["preconditioner"] = preconditioner
    ps["MonodomainSolver"]["default_timestep"] = dt
    ps["MonodomainSolver"]["use_custom_preconditioner"] = False
    ps["theta"] = theta
    ps["enable_adjoint"] = False
    ps["apply_stimulus_current_to_pde"] = True
    # ps["BasicCardiacODESolver"]["scheme"] = scheme
    ps["CardiacODESolver"]["scheme"] = scheme
    # ps["ode_solver_choice"] = "BasicCardiacODESolver"
    # ps["BasicCardiacODESolver"]["V_polynomial_family"] = "CG"
    # ps["BasicCardiacODESolver"]["V_polynomial_degree"] = 1
    # ps["BasicCardiacODESolver"]["S_polynomial_family"] = "CG"
    # ps["BasicCardiacODESolver"]["S_polynomial_degree"] = 1
    return ps


def file_exist(filename: Optional[str], suffix: str) -> bool:
    return (
        filename is not None
        and filename != ""
        and Path(filename).is_file()
        and Path(filename).suffix == suffix
    )


def load_json(filename: str):
    with open(filename, "r") as fid:
        d = json.load(fid)
    return d


def handle_cell_params(
    CellModel: Type[BaseCellModel],
    cell_params: Optional[Dict[str, float]] = None,
    disease_state: str = "healthy",
    drug_factors_file: str = "",
    popu_factors_file: str = "",
):
    cell_params_tmp = CellModel.default_parameters()
    # FIXME: In this case we update the parameters first, while in the
    # initial condition case we do that last. We need to be consistent
    # about this.
    if cell_params is not None:
        cell_params_tmp.update(cell_params)

    CellModel.update_disease_parameters(cell_params_tmp, disease_state=disease_state)
    # Adding optional drug factors to parameters (if drug_factors_file exists)
    if file_exist(drug_factors_file, ".json"):
        logger.info(f"Drug scaling factors loaded from {drug_factors_file}")
        cell_params_tmp.update(load_json(drug_factors_file))
    else:
        if drug_factors_file != "":
            logger.warning(f"Unable to load drug factors file {drug_factors_file}")
    # FIXME: A problem here is that popu_factors_file will overwrite the
    # drug_factors_file. Is it possible to only have one file?
    if file_exist(popu_factors_file, ".json"):
        logger.info(f"Population scaling factors loaded from {popu_factors_file}")
        cell_params_tmp.update(load_json(popu_factors_file))
    else:
        if popu_factors_file != "":
            logger.warning(f"Unable to load popu factors file {popu_factors_file}")

    return cell_params_tmp


def handle_cell_inits(
    CellModel: Type[cbcbeat.CardiacCellModel],
    cell_inits: Optional[Dict[str, float]] = None,
    cell_init_file: str = "",
) -> Dict[str, float]:
    cell_inits_tmp = CellModel.default_initial_conditions()
    if file_exist(cell_init_file, ".json"):
        cell_inits_tmp.update(load_json(cell_init_file))

    if file_exist(cell_init_file, ".h5"):
        cell_inits_tmp.update(load_json(cell_init_file))
        from .save_load_functions import load_initial_conditions_from_h5

        cell_inits = load_initial_conditions_from_h5(
            cell_init_file,
            CellModel=CellModel,
        )

    # FIXME: This is a bit confusing, since it will overwrite the
    # inputs from the cell_init_file. There should be only one way to
    # do this IMO. I think this might be difficult for the user to reason
    # about. I think in general we should handle the loading from files
    # at higher level.
    if cell_inits is not None:
        cell_inits_tmp.update(cell_inits)
    return cell_inits_tmp



from __future__ import annotations

from functools import partial
from pathlib import Path
from typing import Optional
from typing import Tuple
from typing import TYPE_CHECKING
from typing import Union

import dolfin
import numpy as np
from dolfin import FiniteElement  # noqa: F401
from dolfin import MixedElement  # noqa: F401
from dolfin import tetrahedron  # noqa: F401
from dolfin import VectorElement  # noqa: F401

from ... import save_load_functions as io
from ... import utils
from ...config import Config
from ...geometry import BaseGeometry
from ...geometry import load_geometry
from ...time_stepper import TimeStepper
from ..em_model import BaseEMCoupling
from ..em_model import setup_EM_model

if TYPE_CHECKING:
    from ... import mechanics_model
    from ... import datacollector

logger = utils.getLogger(__name__)


class EMCoupling(BaseEMCoupling):
    def __init__(
        self,
        geometry: BaseGeometry,
        **state_params,
    ) -> None:
        super().__init__(geometry=geometry, **state_params)

        self.V_mech = dolfin.FunctionSpace(self.mech_mesh, "CG", 1)
        self.XS_mech = dolfin.Function(self.V_mech, name="XS_mech")
        self.XW_mech = dolfin.Function(self.V_mech, name="XW_mech")

        self.V_ep = dolfin.FunctionSpace(self.ep_mesh, "CG", 1)
        self.lmbda_ep = dolfin.Function(self.V_ep, name="lambda_ep")
        self.Zetas_ep = dolfin.Function(self.V_ep, name="Zetas_ep")
        self.Zetaw_ep = dolfin.Function(self.V_ep, name="Zetaw_ep")

        self.transfer_matrix = dolfin.PETScDMCollection.create_transfer_matrix(
            self.V_mech,
            self.V_ep,
        ).mat()

    def interpolate(
        self,
        f_mech: dolfin.Function,
        f_ep: dolfin.Function,
    ) -> dolfin.Function:
        """Interpolates function from mechanics to ep mesh"""

        x = dolfin.as_backend_type(f_mech.vector()).vec()
        a, temp = self.transfer_matrix.getVecs()

        self.transfer_matrix.mult(x, temp)
        f_ep.vector().vec().aypx(0.0, temp)
        f_ep.vector().apply("")

        # Remember to free memory allocated by petsc: https://gitlab.com/petsc/petsc/-/issues/1309
        x.destroy()
        a.destroy()
        temp.destroy()

    @property
    def coupling_type(self):
        return "fully_coupled_ORdmm_Land"

    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, type(self)):
            return NotImplemented

        if not super().__eq__(__o):
            return False

        for attr in [
            "vs",
            "mech_state",
            "lmbda_mech",
            "Zetas_mech",
            "Zetaw_mech",
            "lmbda_ep",
            "Zetas_ep",
            "Zetaw_ep",
            "XS_mech",
            "XW_mech",
        ]:
            if not np.allclose(
                getattr(self, attr).vector().get_local(),
                getattr(__o, attr).vector().get_local(),
            ):
                logger.info(f"{attr} differs in equality")
                return False

        return True

    def register_time_stepper(self, time_stepper: TimeStepper) -> None:
        super().register_time_stepper(time_stepper)
        self.mech_solver.material.active.register_time_stepper(time_stepper)

    @property
    def dt_mechanics(self) -> float:
        return self.mech_solver.material.active.dt

    @property
    def mech_mesh(self):
        return self.geometry.mechanics_mesh

    @property
    def ep_mesh(self):
        return self.geometry.ep_mesh

    @property
    def mech_state(self) -> dolfin.Function:
        return self.mech_solver.state

    @property
    def vs(self) -> dolfin.Function:
        return self.ep_solver.solution_fields()[0]

    @property
    def assigners(self) -> datacollector.Assigners:
        return self._assigners

    @assigners.setter
    def assigners(self, assigners) -> None:
        self._assigners = assigners

    def setup_assigners(self) -> None:
        from ...datacollector import Assigners

        self.assigners = Assigners(vs=self.vs, mech_state=self.mech_state)
        for name, index in [
            ("V", 0),
            ("Ca", 45),
            ("CaTrpn", 42),
            ("TmB", 43),
            ("Cd", 44),
            ("XS", 40),
            ("XW", 41),
        ]:
            self.assigners.register_subfunction(
                name=name,
                group="ep",
                subspace_index=index,
            )

        self.assigners.register_subfunction(
            name="u",
            group="mechanics",
            subspace_index=self.mech_solver.u_subspace_index,
        )

        for name, index in [
            ("XS", 40),
            ("XW", 41),
        ]:
            self.assigners.register_subfunction(
                name=name,
                group="ep",
                subspace_index=index,
                is_pre=True,
            )

        self.coupling_to_mechanics()

    def register_ep_model(self, solver):
        logger.debug("Registering EP model")
        self.ep_solver = solver

        if hasattr(self, "mech_solver"):
            self.mechanics_to_coupling()
        self.coupling_to_mechanics()
        logger.debug("Done registering EP model")

    def register_mech_model(self, solver: mechanics_model.MechanicsProblem):
        logger.debug("Registering mech model")
        self.mech_solver = solver

        self.Zetas_mech = solver.material.active.Zetas_prev
        self.Zetaw_mech = solver.material.active.Zetaw_prev
        self.lmbda_mech = solver.material.active.lmbda_prev

        # Note sure why we need to do this for the LV?
        self.lmbda_mech.set_allow_extrapolation(True)
        self.Zetas_mech.set_allow_extrapolation(True)
        self.Zetaw_mech.set_allow_extrapolation(True)

        self.mechanics_to_coupling()
        if hasattr(self, "ep_solver"):
            self.coupling_to_mechanics()
        logger.debug("Done registering EP model")

    def update_prev_mechanics(self):
        self.mech_solver.material.active.update_prev()

    def update_prev_ep(self):
        self.ep_solver.vs_.assign(self.ep_solver.vs)

    def ep_to_coupling(self):
        logger.debug("Update mechanics")
        self.assigners.assign_ep()
        logger.debug("Done updating mechanics")

    def coupling_to_mechanics(self):
        logger.debug("Interpolate mechanics")
        if hasattr(self, "_assigners"):
            self.XS_mech.interpolate(self.assigners.functions["ep"]["XS"])
            self.XW_mech.interpolate(self.assigners.functions["ep"]["XW"])
        logger.debug("Done interpolating mechanics")

    def mechanics_to_coupling(self):
        logger.debug("Interpolate EP")
        self.interpolate(self.lmbda_mech, self.lmbda_ep)
        self.interpolate(self.Zetas_mech, self.Zetas_ep)
        self.interpolate(self.Zetaw_mech, self.Zetaw_ep)
        logger.debug("Done interpolating EP")

    def coupling_to_ep(self):
        logger.debug("Update EP")
        logger.debug("Done updating EP")

    def solve_mechanics(self) -> None:
        logger.debug("Solve mechanics")
        self.mech_solver.solve()

    def solve_ep(self, interval: Tuple[float, float]) -> None:
        logger.debug("Solve EP")
        self.ep_solver.step(interval)

    def print_mechanics_info(self):
        total_dofs = self.mech_state.function_space().dim()
        utils.print_mesh_info(self.mech_mesh, total_dofs)
        logger.info("Mechanics model")

    def print_ep_info(self):
        # Output some degrees of freedom
        total_dofs = self.vs.function_space().dim()
        logger.info("EP model")
        utils.print_mesh_info(self.ep_mesh, total_dofs)

    def cell_params(self):
        return self.ep_solver.ode_solver._model.parameters()

    def register_datacollector(self, collector: datacollector.DataCollector) -> None:
        super().register_datacollector(collector=collector)

        collector.register("ep", "Zetas", self.Zetas_ep)
        collector.register("ep", "Zetaw", self.Zetaw_ep)
        collector.register("ep", "lambda", self.lmbda_ep)
        collector.register("mechanics", "XS", self.XS_mech)
        collector.register("mechanics", "XW", self.XW_mech)
        collector.register("mechanics", "Zetas", self.Zetas_mech)
        collector.register("mechanics", "Zetaw", self.Zetaw_mech)
        collector.register("mechanics", "lambda", self.lmbda_mech)
        collector.register(
            "mechanics",
            "Ta",
            self.mech_solver.material.active.Ta_current,
        )
        self.mech_solver.solver.register_datacollector(collector)

    def save_state(
        self,
        path: Union[str, Path],
        config: Optional[Config] = None,
    ) -> None:
        super().save_state(path=path, config=config)

        with dolfin.HDF5File(
            self.geometry.comm(),
            Path(path).as_posix(),
            "a",
        ) as h5file:
            h5file.write(self.lmbda_mech, "/em/lmbda_prev")
            h5file.write(self.Zetas_mech, "/em/Zetas_prev")
            h5file.write(self.Zetaw_mech, "/em/Zetaw_prev")
            h5file.write(self.ep_solver.vs, "/ep/vs")
            h5file.write(self.mech_solver.state, "/mechanics/state")

        io.dict_to_h5(
            self.cell_params(),
            path,
            "ep/cell_params",
        )

    @classmethod
    def from_state(
        cls,
        path: Union[str, Path],
        drug_factors_file: Union[str, Path] = "",
        popu_factors_file: Union[str, Path] = "",
        disease_state="healthy",
        PCL: float = 1000,
    ) -> BaseEMCoupling:
        logger.debug(f"Load state from path {path}")
        path = Path(path)
        if not path.is_file():
            raise FileNotFoundError(f"File {path} does not exist")

        geo = load_geometry(path, schema_path=path.with_suffix(".json"))
        logger.debug("Open file with h5py")
        with io.h5pyfile(path) as h5file:
            config = Config(**io.h5_to_dict(h5file["config"]))
            state_params = io.h5_to_dict(h5file["state_params"])
            cell_params = io.h5_to_dict(h5file["ep"]["cell_params"])
            vs_signature = h5file["ep"]["vs"].attrs["signature"].decode()
            mech_signature = h5file["mechanics"]["state"].attrs["signature"].decode()

        config.drug_factors_file = drug_factors_file
        config.popu_factors_file = popu_factors_file
        config.disease_state = disease_state
        config.PCL = PCL

        VS = dolfin.FunctionSpace(geo.ep_mesh, eval(vs_signature))
        vs = dolfin.Function(VS)

        W = dolfin.FunctionSpace(geo.mechanics_mesh, eval(mech_signature))
        mech_state = dolfin.Function(W)

        # FIXME: load this signature from the file as well
        V = dolfin.FunctionSpace(geo.mechanics_mesh, "CG", 1)
        lmbda_prev = dolfin.Function(V, name="lambda")
        Zetas_prev = dolfin.Function(V, name="Zetas")
        Zetaw_prev = dolfin.Function(V, name="Zetaw")
        logger.debug("Load functions")
        with dolfin.HDF5File(geo.ep_mesh.mpi_comm(), path.as_posix(), "r") as h5file:
            h5file.read(vs, "/ep/vs")
            h5file.read(mech_state, "/mechanics/state")
            h5file.read(lmbda_prev, "/em/lmbda_prev")
            h5file.read(Zetas_prev, "/em/Zetas_prev")
            h5file.read(Zetaw_prev, "/em/Zetaw_prev")

        from . import CellModel, ActiveModel

        cell_inits = io.vs_functions_to_dict(
            vs,
            state_names=CellModel.default_initial_conditions().keys(),
        )

        cls_ActiveModel = partial(
            ActiveModel,
            Zetas=Zetas_prev,
            Zetaw=Zetaw_prev,
            lmbda=lmbda_prev,
        )

        return setup_EM_model(
            cls_EMCoupling=cls,
            cls_CellModel=CellModel,
            cls_ActiveModel=cls_ActiveModel,
            geometry=geo,
            config=config,
            cell_inits=cell_inits,
            cell_params=cell_params,
            mech_state_init=mech_state,
            state_params=state_params,
        )



from pathlib import Path
from unittest import mock

import dolfin
import h5py
import pytest
import simcardems


def test_DataCollector_reset_state_when_file_exists(tmp_path, geo):
    simcardems.DataCollector(tmp_path, geo=geo)

    with mock.patch("simcardems.utils.remove_file") as remove_file_mock:
        collector = simcardems.DataCollector(tmp_path, geo=geo, reset_state=True)
    remove_file_mock.assert_called()
    assert Path(collector.results_file).is_file()


def test_DataCollector_not_reset_state_when_file_exists(tmp_path, geo):
    simcardems.DataCollector(tmp_path, geo=geo)

    with mock.patch("simcardems.utils.remove_file") as remove_file_mock:
        collector = simcardems.DataCollector(tmp_path, geo=geo, reset_state=False)
    remove_file_mock.assert_not_called()
    assert Path(collector.results_file).is_file()


def test_DataCollector_create_file_with_geo(tmp_path, geo):
    collector = simcardems.DataCollector(tmp_path, geo=geo)
    assert Path(collector.results_file).is_file()
    with h5py.File(collector.results_file, "r") as h5file:
        assert "geometry" in h5file


def test_DataCollector_register(tmp_path, geo):
    V = dolfin.FunctionSpace(geo.mesh, "CG", 1)
    f = dolfin.Function(V)

    simcardems.set_log_level(10)
    collector = simcardems.DataCollector(tmp_path, geo=geo)
    collector.register("ep", "func", f)
    assert "func" in collector.names["ep"]

    with pytest.raises(ValueError) as ex:
        simcardems.DataLoader(collector.results_file)
    assert str(ex.value) == "No functions found in results file"


@pytest.mark.parametrize("group", ["ep", "mechanics"])
def test_DataCollector_store(group, geo, tmp_path):
    mesh = geo.mesh if group == "mechanics" else geo.ep_mesh
    V = dolfin.FunctionSpace(mesh, "CG", 1)
    f = dolfin.Function(V)
    f.vector()[:] = 42

    simcardems.set_log_level(10)
    collector = simcardems.DataCollector(tmp_path, geo=geo)
    collector.register(group, "func", f)
    assert "func" in collector.names[group]

    collector.store(0)
    loader = simcardems.DataLoader(collector.results_file)
    assert loader.time_stamps == ["0.00"]

    g1 = loader.get(group, "func", "0.00")
    g2 = loader.get(group, "func", 0)
    assert all(g1.vector().get_local() == f.vector().get_local())
    assert all(g2.vector().get_local() == f.vector().get_local())


def test_DataLoader_load_empty_files_raises_ValueError(tmp_path, geo):
    collector = simcardems.DataCollector(tmp_path, geo=geo)
    with pytest.raises(ValueError):
        simcardems.DataLoader(collector.results_file)


def test_DataCollector_store_version(tmp_path, geo):
    collector = simcardems.DataCollector(tmp_path, geo=geo)
    loader = simcardems.DataLoader(collector.results_file, empty_ok=True)
    from packaging.version import parse

    v = parse(simcardems.__version__)
    assert loader.version_major == v.major
    assert loader.version_minor == v.minor
    assert loader.version_micro == v.micro



from . import cell_model
from . import em_model
from .cell_model import ORdmmLandPureEp as CellModel
from .em_model import EMCoupling

ActiveModel = None
loggers = [
    "simcardems.models.pure_ep_ORdmm_Land.cell_model.logger",
    "simcardems.models.pure_ep_ORdmm_Land.cell_model.em_model.logger",
]

__all__ = [
    "cell_model",
    "em_model",
    "EMCoupling",
    "CellModel",
    "ActiveModel",
    "loggers",
]



from __future__ import annotations

from functools import partial
from pathlib import Path
from typing import Optional
from typing import Tuple
from typing import TYPE_CHECKING
from typing import Union

import cbcbeat
import dolfin
import numpy as np
import ufl
from dolfin import FiniteElement  # noqa: F401
from dolfin import MixedElement  # noqa: F401
from dolfin import tetrahedron  # noqa: F401
from dolfin import VectorElement  # noqa: F401

from .. import em_model
from ... import geometry as _geometry
from ... import save_load_functions as io
from ... import utils
from ...config import Config


if TYPE_CHECKING:
    from ... import datacollector
    from ... import mechanics_model


logger = utils.getLogger(__name__)


def Ta(XS, XW, Zetas, Zetaw, lmbda, Tref, rs, Beta0):
    _min = ufl.min_value
    _max = ufl.max_value
    if isinstance(lmbda, (int, float)):
        _min = min
        _max = max
    lmbda = _min(1.2, lmbda)
    h_lambda_prima = 1 + Beta0 * (lmbda + _min(lmbda, 0.87) - 1.87)
    h_lambda = _max(0, h_lambda_prima)

    return h_lambda * (Tref / rs) * (XS * (Zetas + 1) + XW * Zetaw)


class EMCoupling(em_model.BaseEMCoupling):
    def __init__(
        self,
        geometry: _geometry.BaseGeometry,
        lmbda: Optional[dolfin.Function] = None,
        **state_params,
    ) -> None:
        super().__init__(geometry=geometry, **state_params)

        self.V_mech = dolfin.FunctionSpace(self.mech_mesh, "CG", 1)
        self.XS_mech = dolfin.Function(self.V_mech, name="XS_mech")
        self.XW_mech = dolfin.Function(self.V_mech, name="XW_mech")
        self.Zetas_mech = dolfin.Function(self.V_mech, name="Zetas_mech")
        self.Zetaw_mech = dolfin.Function(self.V_mech, name="Zetaw_mech")
        self._lmbda_mech_func = dolfin.Function(self.V_mech, name="Zetaw_mech")
        self._lmbda_mech_func.set_allow_extrapolation(True)

        self.V_ep = dolfin.FunctionSpace(self.ep_mesh, "CG", 1)
        self.XS_ep = dolfin.Function(self.V_ep, name="XS_ep")
        self.XW_ep = dolfin.Function(self.V_ep, name="XW_ep")
        self.lmbda_ep = dolfin.Function(self.V_ep, name="lambda_ep")
        self.lmbda_ep_prev = dolfin.Function(self.V_ep, name="lambda_ep_prev")
        self._dLambda = dolfin.Function(self.V_ep, name="dLambda")
        self.Zetas_ep = dolfin.Function(self.V_ep, name="Zetas_ep")
        self.Zetaw_ep = dolfin.Function(self.V_ep, name="Zetaw_ep")

        # self.W_ep = dolfin.VectorFunctionSpace(self.ep_mesh, "CG", 2)
        # self.u_ep = dolfin.Function(self.W_ep, name="u_ep")

        self._projector_V_ep = utils.Projector(self.V_ep)
        self._projector_V_mech = utils.Projector(self.V_mech)

        if lmbda is not None:
            self.lmbda_ep.vector()[:] = lmbda.vector()
            self.lmbda_ep_prev.vector()[:] = lmbda.vector()

        self.transfer_matrix = dolfin.PETScDMCollection.create_transfer_matrix(
            self.V_mech,
            self.V_ep,
        ).mat()

    def interpolate(
        self,
        f_mech: dolfin.Function,
        f_ep: dolfin.Function,
    ) -> dolfin.Function:
        """Interpolates function from mechanics to ep mesh"""

        x = dolfin.as_backend_type(f_mech.vector()).vec()
        a, temp = self.transfer_matrix.getVecs()
        self.transfer_matrix.mult(x, temp)
        f_ep.vector().vec().aypx(0.0, temp)
        f_ep.vector().apply("")

        # Remember to free memory allocated by petsc: https://gitlab.com/petsc/petsc/-/issues/1309
        x.destroy()
        a.destroy()
        temp.destroy()

    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, type(self)):
            return NotImplemented

        if not super().__eq__(__o):
            return False

        for attr in [
            "vs",
            "mech_state",
            "lmbda_mech_func",
            "Zetas_mech",
            "Zetaw_mech",
            "lmbda_ep",
            "lmbda_ep_prev",
            "dLambda_ep",
            "Zetas_ep",
            "Zetaw_ep",
            "XS_ep",
            "XW_ep",
            "XS_mech",
            "XW_mech",
        ]:
            if not np.allclose(
                getattr(self, attr).vector().get_local(),
                getattr(__o, attr).vector().get_local(),
            ):
                logger.info(f"{attr} differs in equality")
                return False

        return True

    @property
    def coupling_type(self):
        return "explicit_ORdmm_Land"

    @property
    def assigners(self) -> datacollector.Assigners:
        return self._assigners

    @assigners.setter
    def assigners(self, assigners) -> None:
        self._assigners = assigners

    def setup_assigners(self) -> None:
        from ...datacollector import Assigners

        self.assigners = Assigners(vs=self.vs, mech_state=self.mech_state)
        for name, index in [
            ("V", 0),
            ("Ca", 45),
            ("XS", 40),
            ("XW", 41),
            ("CaTrpn", 42),
            ("TmB", 43),
            ("Cd", 44),
        ]:
            self.assigners.register_subfunction(
                name=name,
                group="ep",
                subspace_index=index,
            )

        self.assigners.register_subfunction(
            name="u",
            group="mechanics",
            subspace_index=self.mech_solver.u_subspace_index,
        )

        for name, index in [
            ("XS", 40),
            ("XW", 41),
        ]:
            self.assigners.register_subfunction(
                name=name,
                group="ep",
                subspace_index=index,
                is_pre=True,
            )

    @property
    def mech_mesh(self):
        return self.geometry.mechanics_mesh

    @property
    def ep_mesh(self):
        return self.geometry.ep_mesh

    @property
    def mech_state(self):
        return self.mech_solver.state

    @property
    def vs(self) -> dolfin.Function:
        return self.ep_solver.solution_fields()[0]

    @property
    def dLambda_ep(self):
        self._dLambda.vector()[:] = self.lmbda_ep.vector() - self.lmbda_ep_prev.vector()
        return self._dLambda

    @property
    def lmbda_mech(self):
        F = dolfin.grad(self.u_mech) + dolfin.Identity(3)
        f = F * self.geometry.f0
        return dolfin.sqrt(f**2)

    @property
    def lmbda_mech_func(self):
        self._projector_V_mech(self._lmbda_mech_func, self.lmbda_mech)
        return self._lmbda_mech_func

    def update_prev_mechanics(self):
        super().update_prev_mechanics()
        # Update previous lmbda
        self.lmbda_ep_prev.vector()[:] = self.lmbda_ep.vector()

    def update_prev_ep(self):
        self.ep_solver.vs_.assign(self.ep_solver.vs)

    # def _project_lmbda(self):
    #     F = dolfin.grad(self.u_ep) + dolfin.Identity(3)
    #     f = F * self.geometry.f0_ep
    #     self._projector_V_ep(self.lmbda_ep, dolfin.project(dolfin.sqrt(f**2)))

    def register_ep_model(self, solver: cbcbeat.SplittingSolver):
        logger.debug("Registering EP model")
        self.ep_solver = solver
        self.XS_ep, self.XS_ep_assigner = utils.setup_assigner(self.vs, 40)
        self.XW_ep, self.XW_ep_assigner = utils.setup_assigner(self.vs, 41)
        self.Zetas_ep, self.Zetas_ep_assigner = utils.setup_assigner(self.vs, 46)
        self.Zetaw_ep, self.Zetaw_ep_assigner = utils.setup_assigner(self.vs, 47)

        params = self.ep_solver.ode_solver._model.parameters()
        self._Tref = params["Tref"]
        self._rs = params["rs"]
        self._Beta0 = params["Beta0"]

        if hasattr(self, "mech_solver"):
            self.mechanics_to_coupling()
            self.coupling_to_mechanics()
        logger.debug("Done registering EP model")

    def register_mech_model(self, solver: mechanics_model.MechanicsProblem):
        logger.debug("Registering mech model")
        self.mech_solver = solver
        self.Ta_mech = self.mech_solver.material.activation

        self._u_subspace_index = (
            1 if type(solver).__name__ == "RigidMotionProblem" else 0
        )
        self.u_mech, self.u_mech_assigner = utils.setup_assigner(
            solver.state,
            self._u_subspace_index,
        )
        self.mechanics_to_coupling()
        if hasattr(self, "ep_solver"):
            self.coupling_to_mechanics()
        logger.debug("Done registering EP model")

    def register_datacollector(self, collector: datacollector.DataCollector) -> None:
        super().register_datacollector(collector=collector)

        collector.register("ep", "lambda", self.lmbda_ep)
        collector.register("ep", "dLambda", self.dLambda_ep)
        collector.register("mechanics", "Ta", self.Ta_mech)
        self.mech_solver.solver.register_datacollector(collector)

    def ep_to_coupling(self):
        logger.debug("Transfer variables from EP to coupling")
        self.XS_ep_assigner.assign(self.XS_ep, utils.sub_function(self.vs, 40))
        self.XW_ep_assigner.assign(self.XW_ep, utils.sub_function(self.vs, 41))
        self.Zetas_ep_assigner.assign(self.Zetas_ep, utils.sub_function(self.vs, 46))
        self.Zetaw_ep_assigner.assign(self.Zetaw_ep, utils.sub_function(self.vs, 47))

        logger.debug("Done transferring variables from EP to coupling")

    def coupling_to_mechanics(self):
        logger.debug("Transfer variables from coupling to mechanics")
        self.XS_mech.interpolate(self.XS_ep)
        self.XW_mech.interpolate(self.XW_ep)
        self.Zetas_mech.interpolate(self.Zetas_ep)
        self.Zetaw_mech.interpolate(self.Zetaw_ep)
        self._projector_V_mech(
            self.Ta_mech,
            Ta(
                XS=self.XS_mech,
                XW=self.XW_mech,
                Zetas=self.Zetas_mech,
                Zetaw=self.Zetaw_mech,
                lmbda=self.lmbda_mech,
                Tref=self._Tref,
                rs=self._rs,
                Beta0=self._Beta0,
            ),
        )
        logger.debug("Done transferring variables from coupling to mechanics")

    def mechanics_to_coupling(self):
        logger.debug("Transfer variables from mechanics to coupling")
        self.u_mech_assigner.assign(
            self.u_mech,
            utils.sub_function(self.mech_state, self._u_subspace_index),
        )
        self.interpolate(self.lmbda_mech_func, self.lmbda_ep)

        logger.debug("Done transferring variables from mechanics to coupling")

    def coupling_to_ep(self):
        logger.debug("Update EP")
        logger.debug("Done updating EP")

    def solve_mechanics(self) -> None:
        self.mech_solver.solve()

    def solve_ep(self, interval: Tuple[float, float]) -> None:
        self.ep_solver.step(interval)

    def print_mechanics_info(self):
        total_dofs = self.mech_state.function_space().dim()
        utils.print_mesh_info(self.mech_mesh, total_dofs)
        logger.info("Mechanics model")

    def print_ep_info(self):
        # Output some degrees of freedom
        total_dofs = self.vs.function_space().dim()
        logger.info("EP model")
        utils.print_mesh_info(self.ep_mesh, total_dofs)

    def cell_params(self):
        return self.ep_solver.ode_solver._model.parameters()

    def save_state(
        self,
        path: Union[str, Path],
        config: Optional[Config] = None,
    ) -> None:
        super().save_state(path=path, config=config)

        with dolfin.HDF5File(
            self.geometry.comm(),
            Path(path).as_posix(),
            "a",
        ) as h5file:
            h5file.write(self.lmbda_ep_prev, "/em/lmbda_prev")
            h5file.write(self.ep_solver.vs, "/ep/vs")
            h5file.write(self.mech_solver.state, "/mechanics/state")

        io.dict_to_h5(
            self.cell_params(),
            path,
            "ep/cell_params",
        )

    @classmethod
    def from_state(
        cls,
        path: Union[str, Path],
        drug_factors_file: Union[str, Path] = "",
        popu_factors_file: Union[str, Path] = "",
        disease_state="healthy",
        PCL: float = 1000,
    ) -> em_model.BaseEMCoupling:
        logger.debug(f"Load state from path {path}")
        path = Path(path)
        if not path.is_file():
            raise FileNotFoundError(f"File {path} does not exist")

        geo = _geometry.load_geometry(path, schema_path=path.with_suffix(".json"))
        logger.debug("Open file with h5py")
        with io.h5pyfile(path) as h5file:
            config = Config(**io.h5_to_dict(h5file["config"]))
            state_params = io.h5_to_dict(h5file["state_params"])
            cell_params = io.h5_to_dict(h5file["ep"]["cell_params"])
            vs_signature = h5file["ep"]["vs"].attrs["signature"].decode()
            mech_signature = h5file["mechanics"]["state"].attrs["signature"].decode()

        config.drug_factors_file = drug_factors_file
        config.popu_factors_file = popu_factors_file
        config.disease_state = disease_state
        config.PCL = PCL

        VS = dolfin.FunctionSpace(geo.ep_mesh, eval(vs_signature))
        vs = dolfin.Function(VS)

        W = dolfin.FunctionSpace(geo.mechanics_mesh, eval(mech_signature))
        mech_state = dolfin.Function(W)

        V = dolfin.FunctionSpace(geo.ep_mesh, "CG", 1)
        lmbda = dolfin.Function(V, name="lambda")
        logger.debug("Load functions")
        with dolfin.HDF5File(geo.ep_mesh.mpi_comm(), path.as_posix(), "r") as h5file:
            h5file.read(vs, "/ep/vs")
            h5file.read(mech_state, "/mechanics/state")
            h5file.read(lmbda, "/em/lmbda_prev")

        from . import CellModel, ActiveModel

        cell_inits = io.vs_functions_to_dict(
            vs,
            state_names=CellModel.default_initial_conditions().keys(),
        )

        cls_EMCoupling = partial(cls, lmbda=lmbda)

        return em_model.setup_EM_model(
            cls_EMCoupling=cls_EMCoupling,
            cls_CellModel=CellModel,
            cls_ActiveModel=ActiveModel,
            geometry=geo,
            config=config,
            cell_inits=cell_inits,
            cell_params=cell_params,
            mech_state_init=mech_state,
            state_params=state_params,
        )



from typing import Dict
from typing import Tuple

import dolfin
import pulse
from cardiac_geometries.geometry import MeshTypes

from .geometry import BaseGeometry


class LeftVentricularGeometry(BaseGeometry):
    @staticmethod
    def default_markers() -> Dict[str, Tuple[int, int]]:
        return {
            "BASE": (5, 2),
            "ENDO": (6, 2),
            "EPI": (7, 2),
        }

    def _default_microstructure(
        self,
        mesh: dolfin.Mesh,
        ffun: dolfin.MeshFunction,
    ) -> pulse.Microstructure:
        from cardiac_geometries import lv_ellipsoid_fibers

        return lv_ellipsoid_fibers.create_microstructure(
            function_space=self.parameters["fiber_space"],
            mesh=mesh,
            ffun=ffun,
            markers=self.markers,
            r_short_endo=self.parameters["r_short_endo"],
            r_short_epi=self.parameters["r_short_epi"],
            r_long_endo=self.parameters["r_long_endo"],
            r_long_epi=self.parameters["r_long_epi"],
            alpha_endo=self.parameters["fibers_angle_endo"],
            alpha_epi=self.parameters["fibers_angle_epi"],
        )

    def _default_ffun(self, mesh: dolfin.Mesh) -> dolfin.MeshFunction:
        raise NotImplementedError

    def _default_mesh(self) -> dolfin.Mesh:
        raise NotImplementedError

    @staticmethod
    def default_parameters():
        return {
            "num_refinements": 1,
            "fiber_space": "Quadrature_3",
            "fibers_angle_endo": -60.0,
            "fibers_angle_epi": 60.0,
            "mesh_type": MeshTypes.lv_ellipsoid.value,
            "mu_apex_endo": -3.141592653589793,
            "mu_apex_epi": -3.141592653589793,
            "mu_base_endo": -1.2722641256100204,
            "mu_base_epi": -1.318116071652818,
            "psize_ref": 3.0,
            "r_long_endo": 17.0,
            "r_long_epi": 20.0,
            "r_short_endo": 7.0,
            "r_short_epi": 10.0,
        }

    def validate(self):
        pass

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(parameters={self.parameters})"



import weakref
from enum import Enum
from pathlib import Path
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import cbcbeat
import dolfin
import h5py
import numpy as np
import pulse
from dolfin import FiniteElement  # noqa: F401
from dolfin import tetrahedron  # noqa: F401
from dolfin import VectorElement  # noqa: F401
from typing_extensions import Literal

from . import utils
from .geometry import BaseGeometry
from .geometry import load_geometry
from .save_load_functions import h5pyfile
from .value_extractor import ValueExtractor

logger = utils.getLogger(__name__)


class DataGroups(Enum):
    ep = "ep"
    mechanics = "mechanics"


def assign_item_is_None(
    f: Optional[dolfin.Function],
    assigner: Optional[dolfin.FunctionAssigner],
    index: Optional[int],
) -> bool:
    if f is None:
        return True
    if assigner is None:
        return True
    if index is None:
        return True
    return False


class Assigners:
    """Helper class to assign subfunctions from EP and Mechanics state"""

    def __init__(
        self,
        *,
        vs: Optional[cbcbeat.SplittingSolver] = None,
        mech_state: Optional[pulse.MechanicsProblem] = None,
    ) -> None:
        self.vs = vs
        self.mech_state = mech_state

        self.functions: Dict[str, Dict[str, dolfin.Function]] = {
            "ep": {},
            "mechanics": {},
        }
        self.assigners: Dict[str, Dict[str, dolfin.FunctionAssigner]] = {
            "ep": {},
            "mechanics": {},
        }
        self.subspace_indices: Dict[str, Dict[str, int]] = {
            "ep": {},
            "mechanics": {},
        }

        self.pre_functions: Dict[str, Dict[str, dolfin.Function]] = {
            "ep": {},
            "mechanics": {},
        }
        self.pre_assigners: Dict[str, Dict[str, dolfin.FunctionAssigner]] = {
            "ep": {},
            "mechanics": {},
        }
        self.pre_subspace_indices: Dict[str, Dict[str, int]] = {
            "ep": {},
            "mechanics": {},
        }

    def assign(self) -> None:
        self.assign_mechanics()
        self.assign_ep()

    def assign_pre(self) -> None:
        self.assign_pre_mechanics()
        self.assign_ep()

    def _get(
        self,
        is_pre: bool,
    ) -> Tuple[
        Dict[str, Dict[str, dolfin.Function]],
        Dict[str, Dict[str, dolfin.FunctionAssigner]],
        Dict[str, Dict[str, int]],
    ]:
        if is_pre:
            return (self.pre_functions, self.pre_assigners, self.pre_subspace_indices)
        else:
            return (self.functions, self.assigners, self.subspace_indices)

    def _assign_mechanics(self, is_pre: bool) -> None:
        functions, assigners, subspace_indices = self._get(is_pre)

        keys = assigners["mechanics"].keys()
        if len(keys) == 0:
            return

        if self.mech_state is None:
            raise RuntimeError(
                "Unable to assign mechanics, no mechanics state registered",
            )
        for name in keys:
            assigner = assigners["mechanics"].get(name)
            index = subspace_indices["mechanics"].get(name)
            f = functions["mechanics"].get(name)
            if assign_item_is_None(f, assigner, index):
                continue

            # Could ass a typeguard here: https://peps.python.org/pep-0647/
            # but this will need python3.10
            assigner.assign(f, utils.sub_function(self.mech_state, index))  # type: ignore

    def assign_pre_mechanics(self) -> None:
        self._assign_mechanics(is_pre=True)

    def assign_mechanics(self) -> None:
        self._assign_mechanics(is_pre=False)

    def _assign_ep(self, is_pre: bool) -> None:
        functions, assigners, subspace_indices = self._get(is_pre)

        keys = assigners["ep"].keys()
        if len(keys) == 0:
            return
        if self.vs is None:
            raise RuntimeError("Unable to assign EP, no EP state registered")

        for name in keys:
            assigner = assigners["ep"].get(name)
            index = subspace_indices["ep"].get(name)
            f = functions["ep"].get(name)
            if assign_item_is_None(f, assigner, index):
                continue

            # Could ass a typeguard here: https://peps.python.org/pep-0647/
            # but this will need python3.10
            assigner.assign(f, utils.sub_function(self.vs, index))  # type: ignore

    def assign_pre_ep(self) -> None:
        self._assign_ep(is_pre=True)

    def assign_ep(self) -> None:
        self._assign_ep(is_pre=False)

    def compute_pre_norm(self):
        norm = 0.0
        hit = False
        for group in ["ep", "mechanics"]:
            for name, pre_func in self.pre_functions[group].items():
                if name in self.functions[group]:
                    hit = True
                    norm += utils.compute_norm(self.functions[group][name], pre_func)

        if not hit:
            # We never computed any norm so we should not interpret
            # the norm as zero, but rather assume that the norm is large
            # so that we actually solve in case we rely on this value as
            # a threshold
            return np.inf
        return norm

    def register_subfunction(
        self,
        name: str,
        group: Literal["ep", "mechanics"],
        subspace_index: int,
        is_pre: bool = False,
    ) -> None:
        func = self.vs if group == "ep" else self.mech_state
        f, assigner = utils.setup_assigner(func, subspace_index)
        functions, assigners, subspace_indices = self._get(is_pre)
        functions[group][name] = f
        assigners[group][name] = assigner
        subspace_indices[group][name] = subspace_index


class InvalidReductionError(RuntimeError):
    pass


class DataCollector:
    def __init__(
        self,
        outdir,
        geo: BaseGeometry,
        outfilename: str = "results.h5",
        reset_state=True,
    ) -> None:
        self.outdir = Path(outdir)
        self._results_file = self.outdir / outfilename
        self.comm = geo.mesh.mpi_comm()  # FIXME: Is this important?
        self._value_extractor = ValueExtractor(geo)

        if reset_state:
            utils.remove_file(self._results_file)

        self._times_stamps = set()
        if not self._results_file.is_file():
            geo.dump(self.results_file)
            from . import __version__
            from packaging.version import parse

            version = parse(__version__)

            with h5pyfile(self._results_file, "a") as f:
                f.create_dataset("version_major", data=version.major)
                f.create_dataset("version_minor", data=version.minor)
                f.create_dataset("version_micro", data=version.micro)

        else:
            try:
                with h5pyfile(self._results_file, "r") as f:
                    self._times_stamps = set(f["ep"]["V"].keys())
            except KeyError:
                pass

        self._functions: Dict[str, Dict[str, dolfin.Function]] = {
            "ep": {},
            "mechanics": {},
        }
        self._reductions: Dict[str, Dict[str, str]] = {
            "ep": {},
            "mechanics": {},
        }

    @property
    def valid_reductions(self) -> List[str]:
        return ["full", "average"] + list(self._value_extractor.boundary.nodes())

    @property
    def results_file(self):
        return self._results_file.as_posix()

    def register(
        self,
        group: str,
        name: str,
        f: dolfin.Function,
        reduction: str = "full",
    ) -> None:
        assert group in [
            "ep",
            "mechanics",
        ], f"Group has to be 'ep' or 'mechanics', got {group}"

        if name in self.names[group]:
            logger.warning(
                f"Warning: {name} in group {group} is already registered - overwriting",
            )
        if reduction not in self.valid_reductions:
            # Check that we have a valid reduction
            raise InvalidReductionError(
                "Invalid reduction {reduction}. Possible options are {}",
            )

        self._functions[group][name] = f
        self._reductions[group][name] = reduction

    @property
    def names(self) -> Dict[str, List[str]]:
        return {k: list(v.keys()) for k, v in self._functions.items()}

    @property
    def function_names(self) -> Dict[str, List[str]]:
        return {
            k: [k1 for k1, v1 in v.items() if v1 is not None]
            for k, v in self._functions.items()
        }

    def store(self, t: float) -> None:
        t_str = f"{t:.2f}"
        logger.debug(f"Store results at time {t_str}")
        if f"{t_str}" in self._times_stamps:
            logger.info(f"Time stamp {t_str} already exist in file")
            return

        self._times_stamps.add(t_str)

        # First do the full values
        with dolfin.HDF5File(self.comm, self.results_file, "a") as h5file:
            for group, names in self.names.items():
                logger.debug(
                    f"Save full states in HDF5File {self.results_file} for group {group}",
                )
                for name in names:
                    logger.debug(f"Save {name}")
                    if self._reductions[group][name] == "full":
                        f = self._functions[group][name]
                        h5file.write(f, f"{group}/{name}/{t_str}")

        # Next do the other reductions
        with h5pyfile(self.results_file, "a") as h5file:
            for group, names in self.names.items():
                logger.debug(
                    f"Save reduced states in HDF5File {self.results_file} for group {group}",
                )
                for name in names:
                    logger.debug(f"Save {name}")
                    if self._reductions[group][name] == "full":
                        continue

                    f = self._functions[group][name]
                    value = self._value_extractor.eval(
                        f,
                        value=self._reductions[group][name],
                    )
                    if f"{group}/{name}" not in h5file:
                        h5file.create_group(f"{group}/{name}")
                    h5file[f"{group}/{name}"].create_dataset(t_str, data=value)

    def save_residual(self, residual, index):
        logger.debug("Save residual")
        with h5pyfile(self._results_file, "a") as h5file:
            if "residual" not in h5file:
                h5file.create_group("residual")
            h5file["residual"].create_dataset(str(index), data=residual)


def close_h5file(h5file):
    if h5file is not None:
        h5file.close()


def close_h5pyfile(h5pyfile):
    if h5pyfile is not None:
        h5pyfile.__exit__()


def extract_number_from_h5py(dataset: Optional[h5py.Dataset]) -> Optional[int]:
    if isinstance(dataset, h5py.Dataset):
        return dataset[...].item()

    return None


class DataLoader:
    def __init__(self, h5name: utils.PathLike, empty_ok: bool = False) -> None:
        self._h5file = None
        self._h5pyfile = None
        self._h5name = Path(h5name)
        if not self._h5name.is_file():
            raise FileNotFoundError(f"File {h5name} does not exist")

        self.geo = load_geometry(self._h5name)
        self._h5pyfile = h5pyfile(self._h5name, "r").__enter__()

        self.version_major = extract_number_from_h5py(
            self._h5pyfile.get("version_major"),
        )
        self.version_minor = extract_number_from_h5py(
            self._h5pyfile.get("version_minor"),
        )
        self.version_micro = extract_number_from_h5py(
            self._h5pyfile.get("version_micro"),
        )

        # Find the remaining functions
        self.names = {
            group: [name for name in self._h5pyfile.get(group, {}).keys()]
            for group in ["ep", "mechanics"]
        }
        if not empty_ok and len(self.names["ep"]) + len(self.names["mechanics"]) == 0:
            raise ValueError("No functions found in results file")

        # Get time stamps
        all_time_stamps = {
            "{group}:{name}": sorted(
                list(self._h5pyfile[group][name].keys()),
                key=lambda x: float(x),
            )
            for group, names in self.names.items()
            for name in names
        }

        # An verify that they are all the same
        self.time_stamps = (
            None
            if not all_time_stamps
            else all_time_stamps[next(iter(all_time_stamps.keys()))]
        )
        for name in all_time_stamps.keys():
            assert self.time_stamps == all_time_stamps[name], name

        if not empty_ok and (self.time_stamps is None or len(self.time_stamps) == 0):
            raise ValueError("No time stamps found")

        # Get the signatures - FIXME: Add a check that the signature exist.
        self._signatures: Dict[str, Dict[str, Optional[str]]] = {}
        for group, names in self.names.items():
            self._signatures[group] = {}
            for name in names:
                try:
                    self._signatures[group][name] = (
                        self._h5pyfile[group][name][self.time_stamps[0]]  # type: ignore
                        .attrs["signature"]
                        .decode()
                    )
                except KeyError:
                    self._signatures[group][name] = None

        if "residual" in self._h5pyfile:
            self._residual = [
                self._h5pyfile["residual"][k][...]
                for k in self._h5pyfile["residual"].keys()
            ]
        else:
            self._residual = []

        self._h5file = dolfin.HDF5File(
            self.ep_mesh.mpi_comm(),
            self._h5name.as_posix(),
            "r",
        )

        self._create_functions()
        self.value_extractor = ValueExtractor(self.geo)
        self._finalizer_h5file = weakref.finalize(self, close_h5file, self._h5file)
        self._finalizer_h5pyfile = weakref.finalize(
            self,
            close_h5pyfile,
            self._h5pyfile,
        )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cleanup()

    def cleanup(self):
        self._finalizer_h5file()
        self._finalizer_h5pyfile()

    @property
    def function_names(self) -> Dict[str, List[str]]:
        return {
            k: [k1 for k1, v1 in v.items() if v1 is not None]
            for k, v in self._functions.items()
        }

    @property
    def residual(self) -> List[np.ndarray]:
        return self._residual

    @property
    def ep_mesh(self):
        return self.geo.ep_mesh

    @property
    def mech_mesh(self):
        return self.geo.mechanics_mesh

    @property
    def size(self) -> int:
        if self.time_stamps is None:
            return 0
        return len(self.time_stamps)

    def __repr__(self):
        return f"{self.__class__.__name__}({str(self._h5name)})"

    def _create_functions(self):
        self._function_spaces = {}

        for group, signature_dict in self._signatures.items():
            mesh = self.ep_mesh if group == "ep" else self.mech_mesh

            self._function_spaces.update(
                {
                    group: {
                        signature: dolfin.FunctionSpace(mesh, eval(signature))
                        for signature in set(signature_dict.values())
                        if signature is not None
                    },
                },
            )

        self._functions = {
            group: {
                name: dolfin.Function(
                    self._function_spaces[group][self._signatures[group][name]],
                )
                for name in names
                if self._signatures[group][name] is not None
            }
            for group, names in self.names.items()
        }

        self._dofs = {
            group: {
                name: func.function_space().tabulate_dof_coordinates()
                for name, func in functions.items()
            }
            for group, functions in self._functions.items()
        }

    def extract_value(
        self,
        group: DataGroups,
        name: str,
        t: Union[str, float],
        reduction: str = "average",
    ):
        func = self.get(group, name, t)
        if isinstance(func, dolfin.Function):
            dofs = self._dofs[self._group_to_str(group)][name]
            return self.value_extractor.eval(func, value=reduction, dofs=dofs)
        return func

    def _group_to_str(self, group):
        group_str = utils.enum2str(group, DataGroups)
        if group_str not in self.names:
            raise KeyError(
                f"Cannot find group {group} in names, expected of of {self.names.keys()}",
            )
        return group_str

    def get(
        self,
        group: DataGroups,
        name: str,
        t: Union[str, float],
    ) -> dolfin.Function:
        """Retrieve the function from the file

        Parameters
        ----------
        group : DataGroups
            The group where the function is stored, either
            'ep' or 'mechanics'
        name : str
            Name of the function
        t : Union[str, float]
            Time stamp you want to use. See `DataLoader.time_stamps`

        Returns
        -------
        dolfin.Function
            The function

        Raises
        ------
        KeyError
            If group does not exist in the file
        KeyError
            If name does not exist in group
        KeyError
            If 't' provided is not among the time stamps
        """
        if self._h5file is None:
            raise RuntimeError(f"Unable to read from file {self._h5name}")

        group_str = self._group_to_str(group)
        names = self.names[group_str]
        if f"{name}" not in names:
            raise KeyError(
                f"Cannot find name {name} in group {group}. Possible options are {names}",
            )

        if isinstance(t, (int, float)):
            t = f"{t:.2f}"

        if self.time_stamps is None:
            raise RuntimeError("No data found in result file")
        if t not in self.time_stamps:
            raise KeyError(f"Invalid time stamps {t}")

        try:
            func = self._functions[group_str][name]
        except KeyError:
            return self._h5pyfile[group_str][name][t][...].item()  # type: ignore
        else:
            self._h5file.read(func, f"{group_str}/{name}/{t}/")
            return func



import json
from collections import defaultdict
from pathlib import Path
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import ap_features as apf
import dolfin
import matplotlib.pyplot as plt
import numpy as np
import tqdm

from . import utils
from .datacollector import DataCollector
from .datacollector import DataGroups
from .datacollector import DataLoader

logger = utils.getLogger(__name__)


def plot_peaks(fname, data, threshold):
    # Find peaks for assessment steady state
    from scipy.signal import find_peaks

    peak_indices = find_peaks(data, height=threshold)

    for i, idx in enumerate(peak_indices[0]):
        if i == 0:
            x = [idx]
            y = [data[idx]]
        else:
            x.append(idx)
            y.append(data[idx])

    # Calculate difference between consecutive list elements
    change_y = [(s - q) / q * 100 for q, s in zip(y, y[1:])]

    fig, ax = plt.subplots()
    ax.plot(change_y)
    ax.set_title("Compare peak values")
    ax.grid()
    ax.set_xlabel("Number of beats")
    ax.set_ylabel("% change from previous beat")
    fig.savefig(fname, dpi=300)


def extract_traces(
    loader: DataLoader,
    reduction: str = "average",
    names: Optional[List[Tuple[str, str]]] = None,
) -> Dict[str, Union[np.ndarray, Dict[str, np.ndarray]]]:
    if loader.time_stamps is None:
        logger.warning("No data found in loader")
        return {}

    values = {
        group: {name: np.zeros(loader.size) for name in names if name != "u"}
        for group, names in loader.names.items()
    }

    if "u" in loader.names.get("mechanics", {}):
        values["mechanics"]["u"] = np.zeros((loader.size, 3))
    values["time"] = np.array(loader.time_stamps, dtype=float)

    logger.info("Extract traces...")

    for group, names_ in loader.names.items():
        logger.info(f"Group: {group}")
        # value_point = getattr(bnd[group], utils.enum2str(point, BoundaryNodes))
        datagroup = getattr(DataGroups, utils.enum2str(group, DataGroups))
        for name in names_:
            if names is not None and (group, name) not in names:
                continue

            logger.info(f"Name: {name}")
            for i, t in enumerate(loader.time_stamps):
                values[group][name][i] = loader.extract_value(
                    datagroup,
                    name,
                    t,
                    reduction=reduction,
                )

    # values["mechanics"]["inv_lmbda"] = 1 - values["mechanics"]["lmbda"]
    return values


def plot_state_traces(
    results_file: utils.PathLike,
    reduction: str = "average",
    flag_peaks: bool = True,
):
    results_file = Path(results_file)
    if not results_file.is_file():
        raise FileNotFoundError(f"File {results_file} does not exist")

    outdir = results_file.parent

    loader = DataLoader(results_file)
    values = extract_traces(loader, reduction=reduction)

    times = np.array(loader.time_stamps, dtype=float)

    if times[-1] > 2000 and flag_peaks:
        plot_peaks(
            outdir.joinpath("compare-peak-values.png"),
            values["ep"]["Ca"],
            0.0002,
        )
    fig, axs = plt.subplots(2, 2, figsize=(10, 8), sharex=True)

    for i, (group, key) in enumerate(
        (("ep", "lambda"), ("mechanics", "Ta"), ("ep", "V"), ("ep", "Ca")),
    ):
        ax = axs.flatten()[i]
        try:
            y = values[group][key]
        except KeyError:
            # Just skip it
            continue
        ax.plot(times, y)
        if key == "lambda":
            ax.set_title(r"$\lambda$")
            ax.set_ylim(min(0.9, min(y)), max(1.1, max(y)))
        else:
            ax.set_title(key)
        ax.grid()

    axs[1, 0].set_xlabel("Time [ms]")
    axs[1, 1].set_xlabel("Time [ms]")
    fig.savefig(outdir.joinpath(f"state_traces_{reduction}.png"), dpi=300)

    fig2, ax2 = plt.subplots(2, 4, figsize=(10, 8), sharex=True)
    for i, (group, key, linestyle, color) in enumerate(
        (
            ("ep", "XS", "solid", "blue"),
            ("ep", "XS", "solid", "blue"),
            ("ep", "CaTrpn", "solid", "blue"),
            ("ep", "TmB", "solid", "blue"),
            ("ep", "Zetas", "solid", "blue"),
            ("ep", "Zetaw", "solid", "blue"),
            ("ep", "Cd", "solid", "blue"),
        ),
    ):
        try:
            y = values[group][key]
        except KeyError:
            # Just skip it
            continue
        ax = ax2.flatten()[i]
        ax.plot(times, y, linestyle=linestyle, color=color)
        ax.set_title(key)
        ax.grid()

    for i in range(4):
        ax2[1, i].set_xlabel("Time [ms]")

    # If there is a residual.txt file: load and plot these results
    if loader.residual:
        # Back to initial dt and time points
        # dt = 0.05  # FIXME: Do we really want to hardcode in this value?
        # times_dt = np.arange(
        #     times[0],
        #     times[0] + len(loader.residual) * dt,
        #     dt,
        # )
        # Change this to linspace so that we know that we have the right
        # number of points
        times_dt = np.linspace(times[0], times[-1], len(loader.residual))

        # First residual
        residual0 = [res[0] for res in loader.residual]
        # Final residual
        residualN = [res[-1] for res in loader.residual]

        ax00r = ax[0, 0].twinx()
        ax00r.plot(
            times_dt,
            residualN,
            "--",
            color="lightcoral",
            label="Newton residualN",
        )
        ax01r = ax[0, 1].twinx()
        ax01r.plot(
            times_dt,
            residualN,
            "--",
            color="lightcoral",
            label="Newton residualN",
        )
        ax01r.set_ylabel("Newton residual N")
        ax10r = ax[1, 0].twinx()
        ax10r.plot(
            times_dt,
            residualN,
            "--",
            color="lightcoral",
            label="Newton residualN",
        )
        ax11r = ax[1, 1].twinx()
        ax11r.plot(
            times_dt,
            residualN,
            "--",
            color="lightcoral",
            label="Newton residualN",
        )
        ax11r.set_ylabel("Newton residual N")
        fig.tight_layout()

        ax200r = ax2[0, 0].twinx()
        ax200r.plot(times_dt, residual0, "--", color="grey", label="Newton residual")
        ax201r = ax2[0, 1].twinx()
        ax201r.plot(times_dt, residual0, "--", color="grey", label="Newton residual")
        ax202r = ax2[0, 2].twinx()
        ax202r.plot(times_dt, residual0, "--", color="grey", label="Newton residual")
        ax203r = ax2[0, 3].twinx()
        ax203r.plot(times_dt, residual0, "--", color="grey", label="Newton residual")
        ax203r.set_ylabel("Newton residual 0")
        ax210r = ax2[1, 0].twinx()
        ax210r.plot(times_dt, residual0, "--", color="grey", label="Newton residual")
        ax211r = ax2[1, 1].twinx()
        ax211r.plot(times_dt, residual0, "--", color="grey", label="Newton residual")
        ax213r = ax2[1, 3].twinx()
        ax213r.plot(times_dt, residual0, "--", color="grey", label="Newton residual")
        ax213r.set_ylabel("Newton residual 0")
        fig2.tight_layout()

        # Create and save figure with Newton residual of first and last iteration
        figNR, axNR = plt.subplots(1, 1)
        axNRr = axNR.twinx()
        axNR.plot(times_dt, residual0, color="grey", label="First res")
        axNRr.plot(times_dt, residualN, color="lightcoral", label="Last residual")
        axNR.set_xlabel("Time (ms)")
        axNR.set_ylabel("Residual 1st Newton iter.")
        axNRr.set_ylabel("Residual last Newton iter.")
        axNR.legend(loc="upper left")
        axNRr.legend(loc="upper right")
        # figNR.tight_layout()
        figNR.savefig(outdir.joinpath("NewtonResidual.png"), dpi=300)

    fig.savefig(outdir.joinpath("state_traces_center.png"), dpi=300)
    fig2.savefig(outdir.joinpath("state_mech_traces_center.png"), dpi=300)


def make_xdmffiles(results_file, names=None):
    loader = DataLoader(results_file)
    outdir = Path(results_file).parent

    for group, _names in loader.function_names.items():
        logger.info(f"Save xdmffile for group {group}")
        for name in _names:
            if names is not None and name not in names:
                continue
            xdmf = dolfin.XDMFFile(
                dolfin.MPI.comm_world,
                outdir.joinpath(f"{group}_{name}.xdmf").as_posix(),
            )
            logger.info(f"Save {name}")
            try:
                for t in tqdm.tqdm(loader.time_stamps):
                    f = loader.get(group, name, t)
                    xdmf.write_checkpoint(
                        f,
                        name,
                        float(t),
                        dolfin.XDMFFile.Encoding.HDF5,
                        True,
                    )

            except RuntimeError:
                logger.info(f"Could not save {name}")


def plot_population(results, outdir, num_models, reset_time=True):
    plt.rcParams["svg.fonttype"] = "none"
    plt.rc("axes", labelsize=13)

    times = np.array(results["m1"]["time"], dtype=float)
    if reset_time:
        times = times - times[0]
    fig, ax = plt.subplots(2, 2, figsize=(10, 8), sharex=True)
    for PoMm in range(1, num_models + 1):
        ax[0, 0].plot(
            times,
            np.array(results[f"m{PoMm}"]["mechanics"]["lambda"], dtype=float),
        )
        ax[0, 1].plot(
            times,
            np.array(results[f"m{PoMm}"]["mechanics"]["Ta"], dtype=float),
        )
        ax[1, 0].plot(times, np.array(results[f"m{PoMm}"]["ep"]["V"], dtype=float))
        ax[1, 1].plot(
            times,
            np.array(results[f"m{PoMm}"]["ep"]["Ca"], dtype=float) * 1000,
        )

    for axi in ax.flatten():
        axi.grid()
        # axi.set_xlim([-20, 1000])
    ax[0, 0].set_ylabel("Active stretch")
    ax[0, 1].set_ylabel("Active tension (kPa)")
    ax[1, 0].set_ylabel("Voltage (mV)")
    ax[1, 0].set_xlabel("Time (ms)")
    ax[1, 1].set_ylabel(r"Intracellular calcium concentration ($\mu$M)")
    ax[1, 1].set_xlabel("Time (ms)")
    fig.savefig(outdir.joinpath("traces_center.png"), dpi=300)
    fig.savefig(outdir.joinpath("traces_center.svg"), format="svg")


def find_duration(y, t, repolarisation):
    s = apf.Beat(y=y, t=t)
    return s.apd(repolarisation)


def find_ttp(y, x):
    s = apf.Beat(y=y, t=x)
    return s.ttp()


def find_decaytime(y, x, a):
    s = apf.Beat(y=y, t=x)
    return s.tau(a / 100)


def stats(y):
    ave = sum(y) / len(y)
    SD = (sum([((x - ave) ** 2) for x in y]) / len(y)) ** 0.5
    return ave, SD


def extract_last_beat(y, time, pacing, return_interval=False):
    allbeats = apf.Beats(y=y, t=time, pacing=pacing)

    try:
        lastbeat = allbeats.beats[-1]

        interval = allbeats.intervals[-1]
    except apf.chopping.EmptyChoppingError:
        lastbeat = allbeats
        interval = (time[0], time[-1])

    start = next(i for i, t in enumerate(time) if t >= interval[0])
    try:
        end = next(i for i, t in enumerate(time) if t >= interval[1])
    except StopIteration:
        end = len(time)

    if return_interval:
        return lastbeat.y, lastbeat.t, (start, end)
    return lastbeat.y, lastbeat.t


def extract_biomarkers(
    *,
    V: np.ndarray,
    time: np.ndarray,
    Ca: np.ndarray,
    Ta: Optional[np.ndarray] = None,
    lmbda: Optional[np.ndarray] = None,
    inv_lmbda: Optional[np.ndarray] = None,
    u: Optional[np.ndarray] = None,
) -> Dict[str, float]:
    d = {}

    V_beat = apf.Beat(y=V, t=time)
    Ca_beat = apf.Beat(y=Ca, t=time)

    d["APD40"] = V_beat.apd(40)
    d["APD50"] = V_beat.apd(50)
    d["APD90"] = V_beat.apd(90)
    d["triangulation"] = V_beat.triangulation(high=90, low=40)
    d["Vpeak"] = np.max(V)
    d["Vmin"] = np.min(V)
    d["dvdt_max"] = V_beat.maximum_upstroke_velocity()
    d["maxCa"] = np.max(Ca)
    d["ampCa"] = np.max(Ca) - np.min(Ca)
    d["CaTD50"] = Ca_beat.apd(50)
    d["CaTD80"] = Ca_beat.apd(80)

    if Ta is not None:
        Ta_beat = apf.Beat(Ta)
        d["maxTa"] = np.max(Ta.y)
        d["ampTa"] = np.max(Ta.y) - np.min(Ta.y)
        d["ttp_Ta"] = Ta_beat.ttp()
        d["rt50_Ta"] = Ta_beat.tau(50)
        d["rt95_Ta"] = Ta_beat.tau(5)
    if lmbda is not None:
        d["maxlmbda"] = np.max(lmbda)
        d["minlmbda"] = np.min(lmbda)

    if inv_lmbda is not None:
        inv_lmbda_beat = apf.Beat(y=inv_lmbda, t=time)
        d["ttplmbda"] = inv_lmbda_beat.ttp()
        d["lmbdaD50"] = inv_lmbda_beat.apd(50)
        d["lmbdaD80"] = inv_lmbda_beat.apd(80)
        d["lmbdaD90"] = inv_lmbda_beat.apd(90)
        d["rt50_lmbda"] = inv_lmbda_beat.tau(50)
        d["rt95_lmbda"] = inv_lmbda_beat.tau(5)

    if u is not None:
        u_norm = np.linalg.norm(u, axis=1)
        ux, uy, uz = u.T
        for name, arr in zip(["norm", "x", "y", "z"], [u_norm, ux, uy, uz]):
            d[f"max_displacement_{name}"] = np.max(arr)
            d[f"min_displacement_{name}"] = np.min(arr)
            d[f"time_to_max_displacement_{name}"] = apf.features.time_to_peak(
                y=arr,
                x=time,
            )
            d[f"time_to_min_displacement_{name}"] = apf.features.time_to_peak(
                y=-arr,
                x=time,
            )

    return d


def get_biomarkers(results, outdir, num_models):
    biomarker_dict = {}
    fig, ax = plt.subplots()
    for PoMm in range(1, num_models + 1):
        biomarker_dict[f"m{PoMm}"] = {}
        # Create numpy arrays for analysis
        time = results[f"m{PoMm}"]["time"]
        V = results[f"m{PoMm}"]["ep"]["V"]
        Ca = results[f"m{PoMm}"]["ep"]["Ca"]
        Ta = results[f"m{PoMm}"]["mechanics"]["Ta"]
        lmbda = results[f"m{PoMm}"]["mechanics"]["lmbda"]
        u = results[f"m{PoMm}"]["mechanics"]["u"]
        inv_lmbda = results[f"m{PoMm}"]["mechanics"]["inv_lmbda"]

        # Select only last beat for further analysis
        onlylastbeat = True
        if onlylastbeat:
            logger.info(
                "Only extracting biomarkers for the last beat of the simulation",
            )
            # Create a list with pacing indicators
            pacing = np.zeros_like(time)
            for i in range(0, len(pacing)):
                if time[i] % 1000 <= 0.09 or i == 0:
                    pacing[i] = 1

            # Overwrite the variables with data from only the last beat
            V, timelb, interval = extract_last_beat(
                V,
                time,
                pacing,
                return_interval=True,
            )
            Ca = Ca[interval[0] : interval[1]]
            Ta = Ta[interval[0] : interval[1]]
            lmbda = lmbda[interval[0] : interval[1]]

            u = u[interval[0] : interval[1], :]
            inv_lmbda = inv_lmbda[interval[0] : interval[1]]
            time = timelb[interval[0] : interval[1]]

            figlast, axlast = plt.subplots()
            axlast.plot(time, V)
            axlast.set_xlabel("Time (ms)")
            axlast.set_ylabel("Voltage (mV)")
            figlast.savefig(outdir.joinpath("AP_lastbeat.png"), dpi=300)

        biomarker_dict[f"m{PoMm}"] = extract_biomarkers(
            V=V,
            Ta=Ta,
            time=time,
            Ca=Ca,
            lmbda=lmbda,
            inv_lmbda=inv_lmbda,
            u=u,
        )

        ax.plot(PoMm, biomarker_dict[f"m{PoMm}"]["APD90"], "*")

    fig.savefig(outdir.joinpath("APD90_permodel.png"), dpi=300)

    with open(outdir.joinpath("biomarkers_PoMcontrol.json"), "w") as f:
        json.dump(biomarker_dict, f)

    # Get average and standard deviation of each biomarker
    biomarker_stats = {}
    for biomarker in biomarker_dict["m1"].keys():
        biomarker_stats[biomarker] = []
        for PoMm in biomarker_dict.keys():
            biomarker_stats[biomarker].append(biomarker_dict[PoMm][biomarker])

        ave, SD = stats(biomarker_stats[biomarker])
        print("Average {} ± SD = {} ± {} ".format(biomarker, ave, SD))

    return biomarker_dict


def json_serial(obj):
    if isinstance(obj, (np.ndarray)):
        return obj.tolist()
    else:
        try:
            return str(obj)
        except Exception:
            raise TypeError("Type %s not serializable" % type(obj))


def numpyfy(d):
    if isinstance(d, (list, tuple)):
        return np.array(d)
    if np.isscalar(d):
        return d

    new_d = {}
    for k, v in d.items():
        new_d[k] = numpyfy(v)
    return new_d


def save_popu_json(population_folder, num_models):
    population_folder = Path(population_folder)
    outfile = population_folder.joinpath("output_dict_center.json")
    if outfile.is_file():
        print("Load json file")
        with open(outfile) as f:
            results = numpyfy(json.load(f))

        if len(results.keys()) != num_models:
            raise Exception(
                "Number of models in json-file is not equal to number of models in cli",
            )
    else:
        results = {}
        for PoMm in range(1, num_models + 1):
            print(f"Analyzing model {PoMm}")
            results_file = population_folder.joinpath("results.h5")
            if not results_file.is_file():
                results_file = population_folder.joinpath(f"m{PoMm}/results.h5")
                if not results_file.is_file():
                    raise FileNotFoundError(f"File {results_file} does not exist")

            loader = DataLoader(results_file)
            results[f"m{PoMm}"] = extract_traces(loader=loader)

        # Save entire dict to json file in outdir(=population_folder)
        with open(outfile, "w") as f:
            json.dump(results, f, indent=2, default=json_serial)

    print("Start plotting")
    plot_population(results, population_folder, num_models)

    print("Start analysis of single node results")
    get_biomarkers(results, population_folder, num_models)


def activation_map(
    voltage: Iterable[dolfin.Function],
    time_stamps: Iterable[float],
    V: dolfin.FunctionSpace,
    threshold: float = 0.0,
    t0: float = 0.0,
) -> dolfin.Function:
    activation_map = dolfin.Function(V)
    activation_map.vector()[:] = t0 - 1

    for t, v in zip(time_stamps, voltage):
        dofs = np.logical_and(
            v.vector()[:] >= threshold,
            activation_map.vector()[:] < t0,
        )

        activation_map.vector()[np.where(dofs)[0]] = float(t)

    return activation_map


def extract_sub_results(
    results_file: utils.PathLike,
    output_file: utils.PathLike,
    t_start: float = 0.0,
    t_end: float | None = None,
    names: Optional[Dict[str, List[str]]] = None,
) -> DataCollector:
    """Extract sub results from another results file.
    This can be useful if you have stored a lot of data in one file
    and you want to create a smaller file containing only a subset
    of the data (e.g only the last beat)

    Parameters
    ----------
    results_file : utils.PathLike
        The input result file
    output_file : utils.PathLike
        Path to file where you want to store the sub results
    t_start : float, optional
        Time point indicating when the sub results should start, by default 0.0
    t_end : float | None, optional
        Time point indicating when the sub results should end, by default None
        in which case it will choose the last time point
    names : Optional[Dict[str, List[str]]], optional
        A dictionary of names for each group indicating which
        functions to extract for the sub results. If not provided (default)
        then all functions will be extracted.

    Returns
    -------
    datacollector.DataCollector
        A data collector containing the sub results

    Raises
    ------
    FileNotFoundError
        If the input file does not exist
    KeyError
        If some of the names provided does not exists
        in the input file
    """
    results_file = Path(results_file)
    if not results_file.is_file():
        raise FileNotFoundError(f"File {results_file} does not exist")

    loader = DataLoader(results_file)
    assert loader.time_stamps is not None
    if names is None:
        # Extract everything
        names = loader.names

    t_start_idx = next(
        (i for i, t in enumerate(map(float, loader.time_stamps)) if t > t_start - 1e-12)
    )
    if t_end is None:
        t_end_idx = len(loader.time_stamps) - 1
    else:
        try:
            t_end_idx = next(
                i
                for i, t in enumerate(map(float, loader.time_stamps))
                if t > t_end + 1e-12
            )
        except StopIteration:
            t_end_idx = len(loader.time_stamps) - 1

    out = Path(output_file)
    collector = DataCollector(
        outdir=out.parent,
        outfilename=out.name,
        geo=loader.geo,
    )

    functions: Dict[str, Dict[str, dolfin.Function]] = defaultdict(dict)
    for group_name, group in names.items():
        for func_name in group:
            try:
                functions[group_name][func_name] = loader._functions[group_name][
                    func_name
                ]
            except KeyError as e:
                raise KeyError(
                    f"Invalid group {group_name} and function {func_name}",
                ) from e
            collector.register(group_name, func_name, functions[group_name][func_name])

    for ti in loader.time_stamps[t_start_idx:t_end_idx]:
        for group_name, group in names.items():
            for func_name in group:
                functions[group_name][func_name].assign(
                    loader.get(DataGroups[group_name], func_name, ti),
                )
        collector.store(float(ti))
    return collector



import json
import math

import pytest
import simcardems
from click.testing import CliRunner
from simcardems.cli import postprocess
from simcardems.cli import run
from simcardems.cli import run_json

mesh_args = ["-lx", 1, "-ly", 1, "-lz", 1, "-dx", 1, "--num_refinements", 1]


@pytest.mark.slow
@pytest.mark.parametrize("coupling_type", simcardems.models.list_coupling_types())
def test_run(coupling_type, tmp_path, geo):
    geometry_path = tmp_path / "geo.h5"
    geometry_schema_path = geometry_path.with_suffix(".json")
    geo.dump(fname=geometry_path, schema_path=geometry_schema_path)
    geo_args = [geometry_path.as_posix(), "-s", geometry_schema_path.as_posix()]

    runner = CliRunner()
    outdir = tmp_path / "results"

    T = 1.0
    arguments = geo_args + [
        "-T",
        T,
        "--outdir",
        outdir.as_posix(),
        "--coupling-type",
        coupling_type,
    ]

    result = runner.invoke(run, arguments)

    assert result.exit_code == 0
    assert outdir.exists()
    assert outdir.joinpath("results.h5").is_file()

    with open(outdir.joinpath("parameters.json"), "r") as f:
        dumped_arguments = json.load(f)

    assert dumped_arguments["geometry_path"] == geometry_path.as_posix()
    assert dumped_arguments["geometry_schema_path"] == geometry_schema_path.as_posix()
    assert math.isclose(dumped_arguments["T"], T)
    assert dumped_arguments["outdir"] == outdir.as_posix()
    assert outdir.joinpath("state.h5").is_file()

    # Make sure we can restart the simulation from the current time point
    T = 2.0
    arguments = geo_args + [
        "-T",
        T,
        "--outdir",
        outdir.as_posix(),
        "--load_state",
    ]

    result = runner.invoke(run, arguments)
    assert result.exit_code == 0
    coupling = getattr(simcardems.models, coupling_type).EMCoupling.from_state(
        path=outdir.joinpath("state.h5"),
    )
    assert math.isclose(coupling.t, 2)
    assert coupling.coupling_type == coupling_type

    # Make sure we can plot the state traces
    result = runner.invoke(postprocess, outdir.as_posix(), "--plot-state-traces")
    assert result.exit_code == 0


@pytest.mark.slow
def test_run_json(tmp_path, geo):
    geometry_path = tmp_path / "geo.h5"
    geometry_schema_path = geometry_path.with_suffix(".json")
    geo.dump(fname=geometry_path, schema_path=geometry_schema_path)
    # geo_args = [geometry_path.as_posix(), "-s", geometry_schema_path.as_posix()]

    runner = CliRunner()
    outdir = tmp_path / "results"

    T = 0.2
    # arguments = geo_args + ["-T", 0.2, "--outdir", outdir.as_posix()]
    # Do not provide schema path to test that it picks up the default schema path
    kwargs = {
        "geometry_path": geometry_path.as_posix(),
        "outdir": outdir.as_posix(),
        "T": T,
    }
    json_path = "args.json"

    with open(json_path, "w") as jsonfile:
        json.dump(kwargs, jsonfile)
    result = runner.invoke(run_json, [json_path])
    assert result.exit_code == 0
    assert outdir.exists()
    assert outdir.joinpath("results.h5").is_file()
    with open(outdir.joinpath("parameters.json"), "r") as f:
        dumped_arguments = json.load(f)
    assert dumped_arguments["geometry_path"] == geometry_path.as_posix()
    assert dumped_arguments["geometry_schema_path"] == geometry_schema_path.as_posix()
    assert math.isclose(dumped_arguments["T"], T)
    assert dumped_arguments["outdir"] == outdir.as_posix()



import math
from pathlib import Path

import dolfin
import numpy as np
import simcardems


def test_activation_map():
    mesh = dolfin.UnitCubeMesh(5, 5, 5)
    V = dolfin.FunctionSpace(mesh, "CG", 1)
    v = dolfin.Function(V)
    t = dolfin.Constant(0.0)

    time_stamps = range(5)

    def voltage_gen():
        for ti in time_stamps:
            t.assign(ti)
            v.interpolate(dolfin.Expression("t * (x[0] + 1) - 4.0", t=t, degree=1))
            yield v

    with dolfin.XDMFFile(dolfin.MPI.comm_world, "volt.xdmf") as volt_file:
        for ti, vi in enumerate(voltage_gen()):
            volt_file.write_checkpoint(
                vi,
                "voltage",
                ti,
                dolfin.XDMFFile.Encoding.HDF5,
                True,
            )

    act = simcardems.postprocess.activation_map(
        voltage=voltage_gen(),
        time_stamps=time_stamps,
        V=V,
        threshold=0.0,
    )

    assert math.isclose(act(0.0, 0.0, 0.0), 4)
    assert math.isclose(act(0.0, 0.0, 0.5), 4)
    assert math.isclose(act(0.5, 0.0, 0.0), 3.0)
    assert math.isclose(act(0.5, 0.5, 0.5), 3.0)
    assert math.isclose(act(1.0, 0.0, 0.0), 2.0)
    assert math.isclose(act(1.0, 0.5, 0.5), 2.0)


def test_extract_sub_results(geo, tmp_path):
    results_file = tmp_path / "results.h5"
    collector = simcardems.DataCollector(
        outdir=results_file.parent,
        outfilename=results_file.name,
        geo=geo,
    )

    # Setup a two mech functions and one ep function
    V_mech = dolfin.FunctionSpace(geo.mesh, "Lagrange", 1)
    f1_mech = dolfin.Function(V_mech)
    collector.register("mechanics", "func1", f1_mech)
    f2_mech = dolfin.Function(V_mech)
    collector.register("mechanics", "func2", f2_mech)

    V_ep = dolfin.FunctionSpace(geo.mesh, "Lagrange", 1)
    f3_ep = dolfin.Function(V_ep)
    collector.register("ep", "func3", f3_ep)

    times = np.arange(0, 10, 0.5)
    for t in times:
        f1_mech.vector()[:] = t
        f2_mech.vector()[:] = 10 + t
        f3_ep.vector()[:] = 42 + t
        collector.store(t)

    loader = simcardems.DataLoader(collector.results_file)

    assert loader.time_stamps == [f"{ti:.2f}" for ti in times]
    sub_results_file = tmp_path / "sub_results.h5"
    sub_collector = simcardems.postprocess.extract_sub_results(
        results_file=results_file,
        output_file=sub_results_file,
        t_start=5.0,
        t_end=7.0,
        names={"mechanics": ["func1"]},
    )

    assert str(sub_collector.results_file) == str(sub_results_file)
    assert Path(sub_results_file).is_file()
    sub_loader = simcardems.DataLoader(sub_results_file)
    assert sub_loader.names == {"ep": [], "mechanics": ["func1"]}
    assert sub_loader.time_stamps == ["5.00", "5.50", "6.00", "6.50", "7.00"]



import datetime
import json
import re
import shutil
import subprocess
import time
from collections import Counter
from pathlib import Path
from textwrap import dedent
from typing import Any
from typing import Dict
from typing import Literal
from typing import Set
from typing import Tuple

import click
import matplotlib.pyplot as plt
import numpy as np
from simcardems.postprocess import extract_biomarkers
from simcardems.postprocess import extract_traces

t0 = time.perf_counter()
import simcardems  # noqa: E402

import_time = time.perf_counter() - t0


def underscore_to_space(x: str) -> str:
    return " ".join(x.split("_"))


def to_string(x):
    if isinstance(x, str):
        return x
    return f"{x:.7g}"


def trace_path(outdir: str, dx: float, dt: float) -> Path:
    return Path(outdir) / f"results_dx{int(100*dx)}_dt{int(1000*dt)}.npy"


def feature_path(outdir: str, dx: float, dt: float) -> Path:
    return Path(outdir) / f"results_dx{int(100*dx)}_dt{int(1000*dt)}.json"


def find_dx_dt(
    outdir: str,
) -> Set[Tuple[float, float]]:
    npy_pattern = re.compile(r"results_dx(?P<dx>\d+)_dt(?P<dt>\d+).npy")
    json_pattern = re.compile(r"results_dx(?P<dx>\d+)_dt(?P<dt>\d+).json")

    npy_dx_dt = set()
    json_dx_dt = set()

    for fname in Path(outdir).iterdir():
        if fname.suffix == ".json":
            m = json_pattern.match(fname.name)
            if m is None:
                continue

            json_dx_dt.add((int(m["dx"]) / 100, int(m["dt"]) / 1000))

        elif fname.suffix == ".npy":
            m = npy_pattern.match(fname.name)
            if m is None:
                continue

            npy_dx_dt.add((int(m["dx"]) / 100, int(m["dt"]) / 1000))

        else:
            continue

    return npy_dx_dt & json_dx_dt


@click.command()
@click.argument(
    "outdir",
    required=True,
    type=click.Path(),
)
@click.option("--dt", type=float, default=0.05, help="Time step used in the EP solver")
@click.option(
    "--dx",
    type=float,
    default=0.2,
    help="Spatial discretization for the mechanics mesh",
)
@click.option("--sha", type=str, default="", help="Current git SHA")
@click.option(
    "--pure-ep",
    is_flag=True,
    default=False,
    help="Whether to run the fully coupled EM model or a pure ep model",
)
def run(outdir: str, dt: float, dx: float, sha, pure_ep):
    coupling_type: Literal["fully_coupled_ORdmm_Land", "pureEP_ORdmm_Land"] = (
        "pureEP_ORdmm_Land" if pure_ep else "fully_coupled_ORdmm_Land"
    )

    data: Dict[str, Any] = {}
    data["import_time"] = import_time
    data["timestamp"] = datetime.datetime.now().isoformat()
    data["simcardems_version"] = simcardems.__version__
    data["dt"] = dt
    data["dx"] = dx
    data["sha"] = sha
    data["coupling_type"] = coupling_type

    config = simcardems.Config(
        outdir=outdir,
        T=1000,
        load_state=True,
        dt=dt,
        coupling_type=coupling_type,
        outfilename="benchmark_results.h5",
    )

    t0 = time.perf_counter()
    geo = simcardems.slabgeometry.SlabGeometry(parameters={"dx": dx})

    if not pure_ep:
        data["num_cells_mechanics"] = geo.mesh.num_cells()
        data["num_vertices_mechanics"] = geo.mesh.num_vertices()

    data["num_cells_ep"] = geo.ep_mesh.num_cells()
    data["num_vertices_ep"] = geo.ep_mesh.num_vertices()

    coupling = simcardems.models.em_model.setup_EM_model_from_config(
        config=config,
        geometry=geo,
    )
    runner = simcardems.Runner.from_models(config=config, coupling=coupling)
    data["create_runner_time"] = time.perf_counter() - t0

    t0 = time.perf_counter()

    if runner.t0 < config.T:
        runner.solve(T=config.T, save_freq=config.save_freq, show_progress_bar=False)
    data["solve_time"] = time.perf_counter() - t0

    loader = simcardems.DataLoader(Path(outdir) / "benchmark_results.h5")

    values = extract_traces(loader=loader)

    kwargs = {
        "V": values["ep"]["V"],
        "time": values["time"],
        "Ca": values["ep"]["Ca"],
    }
    if not pure_ep:
        kwargs.update(
            {
                "Ta": values["mechanics"]["Ta"],
                "lmbda": values["mechanics"]["lambda"],
                "inv_lmbda": 1 - values["mechanics"]["lambda"],
                "u": values["mechanics"]["u"],
            },
        )

    data.update(
        extract_biomarkers(**kwargs),
    )

    with open(feature_path(outdir, dx, dt), "w") as f:
        json.dump(data, f, indent=2)

    np.save(trace_path(outdir, dx, dt), kwargs)

    return 0


@click.command()
@click.argument(
    "outdir",
    required=True,
    type=click.Path(),
)
@click.pass_context
def pure_ep_report(ctx, outdir):
    for dx in [0.1, 0.2, 0.4]:
        ctx.invoke(run, outdir=outdir, dt=0.05, dx=dx, sha="", pure_ep=True)
    for dt in [0.025, 0.1]:
        ctx.invoke(run, outdir=outdir, dt=dt, dx=0.2, sha="", pure_ep=True)

    ctx.invoke(generate_report, outdir=outdir)


def create_table_dx(feature_keys, features, fixed_dt, dxs):
    body = ""
    for feature in feature_keys:
        body += (
            f" {underscore_to_space(feature)} & "
            + " & ".join(
                [
                    underscore_to_space(to_string(features[(dx, fixed_dt)][feature]))
                    for dx in dxs
                ],
            )
            + "\\\\ \n"
        )

    table = dedent(
        """
    \\begin{{center}}
    \\begin{{tabular}}{{ {c} }}
    {heading}
    \\hline
    {body}
    \\end{{tabular}}
    \\end{{center}}
    """,
    ).format(
        c="c" * (len(dxs) + 1),
        heading=" &".join(["feature"] + [f"dt={dx}" for dx in dxs]) + "\\\\ \n",
        body=body,
    )
    return table


def create_table_dt(feature_keys, features, fixed_dx, dts):
    body = ""
    for feature in feature_keys:
        body += (
            f" {underscore_to_space(feature)} & "
            + " & ".join(
                [
                    underscore_to_space(to_string(features[(fixed_dx, dt)][feature]))
                    for dt in dts
                ],
            )
            + "\\\\ \n"
        )

    table = dedent(
        """
    \\begin{{center}}
    \\begin{{tabular}}{{ {c} }}
    {heading}
    \\hline
    {body}
    \\end{{tabular}}
    \\end{{center}}
    """,
    ).format(
        c="c" * (len(dts) + 1),
        heading=" &".join(["feature"] + [f"dt={dt}" for dt in dts]) + "\\\\ \n",
        body=body,
    )
    return table


@click.command()
@click.argument(
    "outdir",
    required=True,
    type=click.Path(),
)
def generate_report(outdir):
    dxs_dts = find_dx_dt(outdir)

    traces = {}
    features = {}

    for dx, dt in dxs_dts:
        traces[(dx, dt)] = np.load(
            trace_path(outdir, dx=dx, dt=dt),
            allow_pickle=True,
        ).item()
        features[(dx, dt)] = json.loads(feature_path(outdir, dx=dx, dt=dt).read_text())

    # Find most frequent dx and dt
    fixed_dx = Counter(map(lambda x: x[0], dxs_dts)).most_common()[0][0]
    fixed_dt = Counter(map(lambda x: x[1], dxs_dts)).most_common()[0][0]

    dxs = sorted([dx for dx, dt in dxs_dts if np.isclose(dt, fixed_dt)])
    dts = sorted([dt for dx, dt in dxs_dts if np.isclose(dx, fixed_dx)])

    keys = [key for key in traces[tuple(traces.keys())[0]].keys() if key != "time"]
    feature_keys = [
        k
        for k in features[tuple(features.keys())[0]].keys()
        if k
        not in ["dt", "dx", "sha", "timestamp", "simcardems_version", "coupling_type"]
    ]

    timestamp = features[tuple(features.keys())[0]]["timestamp"]
    version = features[tuple(features.keys())[0]]["simcardems_version"]
    coupling_type = underscore_to_space(
        features[tuple(features.keys())[0]]["coupling_type"],
    )
    fig, ax = plt.subplots(len(keys), 1, sharex=True)

    for dx in dxs:
        d = traces[(dx, fixed_dt)]
        for i, name in enumerate(keys):
            ax[i].plot(d["time"], d[name], label=rf"$\Delta x = {dx}$")

    for i, name in enumerate(keys):
        ax[i].set_ylabel(name)
        ax[i].legend()
    ax[0].set_title("Spatial convergence")
    fig.savefig(Path(outdir) / "spatial_convergence.png", bbox_inches="tight")

    fig, ax = plt.subplots(len(keys), 1, sharex=True)

    for dt in dts:
        d = traces[(fixed_dx, dt)]
        for i, name in enumerate(keys):
            ax[i].plot(d["time"], d[name], label=rf"$\Delta t = {dt}$")

    for i, name in enumerate(keys):
        ax[i].set_ylabel(name)
        ax[i].legend()

    ax[0].set_title("Temporal convergence")
    fig.savefig(Path(outdir) / "temporal_convergence.png", bbox_inches="tight")

    text = dedent(
        f"""
    \\documentclass{{article}}
    \\usepackage{{graphicx}}
    \\begin{{document}}

    \\section{{Convergence}}
    \\begin{{itemize}}
     \\item Timestamp: {timestamp}
     \\item Simcardems version: {version}
     \\item Model: {coupling_type}
    \\end{{itemize}}

    \\subsection{{Spatial convergence}}
    {create_table_dx(feature_keys, features, fixed_dt, dxs)}
    \\includegraphics{{spatial_convergence.png}}



    \\subsection{{Temporal convergence}}
    {create_table_dt(feature_keys, features, fixed_dx, dts)}
    \\includegraphics{{temporal_convergence.png}}

    \\end{{document}}
    """,
    )

    (Path(outdir) / "report.tex").write_text(text)

    pdflatex = shutil.which("pdflatex")
    # if pdflatex is None:
    print("Unable to generate pdf report since pdflatex is not install")

    msg = """
    export DEBIAN_FRONTEND=noninteractive
    apt-get update &&
    apt-get install texlive-latex-base
        texlive-fonts-recommended
        texlive-fonts-extra
        texlive-latex-extra
    """
    print(f"Try {msg}")

    subprocess.run([pdflatex, "report.tex"], cwd=outdir)


@click.group()
def main():
    pass


main.add_command(run)
main.add_command(pure_ep_report)
main.add_command(generate_report)

if __name__ == "__main__":
    raise SystemExit(main())



import json
import shutil
from pathlib import Path

import numpy as np
import pytest
from simcardems import Config
from simcardems import Runner
from simcardems import TimeStepper


@pytest.mark.slow
def test_runner_load_state_with_new_parameters(tmp_path, geo):
    geometry_path = tmp_path / "geo.h5"
    geometry_schema_path = geometry_path.with_suffix(".json")
    geo.dump(fname=geometry_path, schema_path=geometry_schema_path)

    outdir = Path("runner_load_state_with_new_parameters")
    if outdir.exists():
        shutil.rmtree(outdir)
    runner = Runner(
        Config(
            outdir=outdir,
            geometry_path=geometry_path,
            geometry_schema_path=geometry_schema_path,
        ),
    )
    runner.solve(0.02)

    drug_factors_file = Path("drug_factors_file.json")
    drug_factors = {"scale_drug_INa": 42.43}
    drug_factors_file.write_text(json.dumps(drug_factors))

    popu_factors_file = Path("popu_factors_file.json")
    popu_factors = {"scale_popu_GNa": 13.13}
    popu_factors_file.write_text(json.dumps(popu_factors))

    runner2 = Runner(
        Config(
            outdir=outdir,
            drug_factors_file=drug_factors_file,
            popu_factors_file=popu_factors_file,
            load_state=True,
        ),
    )
    assert np.isclose(
        runner2.coupling.cell_params()["scale_drug_INa"],
        drug_factors["scale_drug_INa"],
    )
    assert np.isclose(
        runner2.coupling.cell_params()["scale_popu_GNa"],
        popu_factors["scale_popu_GNa"],
    )

    drug_factors_file.unlink()
    popu_factors_file.unlink()
    shutil.rmtree(outdir)


def test_time_stepper_ns():
    time_stepper = TimeStepper(t0=0, T=0.2, dt=0.1, use_ns=True)
    assert np.allclose(tuple(time_stepper), ((0, 100000), (100000, 200000)))


def test_time_stepper_ms():
    time_stepper = TimeStepper(t0=0, T=0.2, dt=0.1, use_ns=False)
    assert np.allclose(tuple(time_stepper), ((0, 0.1), (0.1, 0.2)))



from __future__ import annotations

import typing
from pathlib import Path

import cbcbeat
import dolfin
import pulse

from .. import ep_model
from .. import geometry as _geometry
from .. import mechanics_model
from .. import save_load_functions as io
from .. import utils
from ..config import Config
from ..time_stepper import TimeStepper

if typing.TYPE_CHECKING:
    from .. import datacollector

logger = utils.getLogger(__name__)


def setup_EM_model(
    cls_EMCoupling,
    cls_CellModel,
    cls_ActiveModel,
    geometry: _geometry.BaseGeometry,
    config: typing.Optional[Config] = None,
    cell_params: typing.Optional[typing.Dict[str, float]] = None,
    cell_inits: typing.Optional[dolfin.Function] = None,
    mech_state_init: typing.Optional[dolfin.Function] = None,
    state_params: typing.Optional[typing.Dict[str, float]] = None,
) -> BaseEMCoupling:
    if config is None:
        config = Config()

    if state_params is None:
        state_params = {}
    coupling = cls_EMCoupling(geometry, **state_params)

    cellmodel = ep_model.setup_cell_model(
        cls=cls_CellModel,
        coupling=coupling,
        cell_init_file=config.cell_init_file,
        drug_factors_file=config.drug_factors_file,
        popu_factors_file=config.popu_factors_file,
        disease_state=config.disease_state,
        cell_inits=cell_inits,
        cell_params=cell_params,
    )

    # Set-up solver and time it
    solver = ep_model.setup_solver(
        coupling=coupling,
        dt=config.dt,
        PCL=config.PCL,
        cellmodel=cellmodel,
    )
    coupling.register_ep_model(solver)

    mech_heart = mechanics_model.setup_solver(
        coupling=coupling,
        bnd_rigid=config.bnd_rigid,
        pre_stretch=config.pre_stretch,
        traction=config.traction,
        spring=config.spring,
        fix_right_plane=config.fix_right_plane,
        set_material=config.set_material,
        use_custom_newton_solver=config.mechanics_use_custom_newton_solver,
        debug_mode=config.debug_mode,
        ActiveModel=cls_ActiveModel,
    )
    if mech_state_init is not None:
        mech_heart.state.assign(mech_state_init)
    coupling.register_mech_model(mech_heart)
    coupling.setup_assigners()

    return coupling


def setup_EM_model_from_config(
    config: Config,
    geometry: typing.Optional[_geometry.BaseGeometry] = None,
    state_params: typing.Optional[typing.Dict[str, float]] = None,
) -> BaseEMCoupling:
    if geometry is None:
        geometry = _geometry.load_geometry(
            mesh_path=config.geometry_path,
            schema_path=config.geometry_schema_path,
        )

    if config.coupling_type == "explicit_ORdmm_Land":
        from .explicit_ORdmm_Land import EMCoupling, CellModel, ActiveModel
    elif config.coupling_type == "fully_coupled_ORdmm_Land":
        from .fully_coupled_ORdmm_Land import EMCoupling, CellModel, ActiveModel  # type: ignore
    elif config.coupling_type == "pureEP_ORdmm_Land":
        from .pureEP_ORdmm_Land import EMCoupling, CellModel, ActiveModel  # type: ignore
    elif config.coupling_type == "fully_coupled_Tor_Land":
        from .fully_coupled_Tor_Land import EMCoupling, CellModel, ActiveModel  # type: ignore
    else:
        raise ValueError(f"Invalid coupling type: {config.coupling_type}")

    return setup_EM_model(
        cls_EMCoupling=EMCoupling,
        cls_CellModel=CellModel,
        cls_ActiveModel=ActiveModel,
        geometry=geometry,
        config=config,
        state_params=state_params,
    )


class BaseEMCoupling:
    def __init__(
        self, geometry: _geometry.BaseGeometry, t: float = 0.0, **kwargs
    ) -> None:
        logger.debug("Create EM coupling")
        self.geometry = geometry
        self.t = t

    @property
    def state_params(self):
        return {"t": self.t}

    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, type(self)):
            return NotImplemented

        if self.coupling_type != __o.coupling_type:
            return False

        if self.t != __o.t:
            return False

        return True

    def register_time_stepper(self, time_stepper: TimeStepper) -> None:
        self._time_stepper = time_stepper
        self._t_mechanics_prev = time_stepper.t

    @property
    def dt_mechanics(self) -> float:
        return TimeStepper.ns2ms(self._time_stepper.t - self._t_mechanics_prev)

    @property
    def coupling_type(self):
        "CustomType"

    def register_ep_model(self, solver: cbcbeat.SplittingSolver) -> None:
        pass

    def register_mech_model(self, solver: pulse.MechanicsProblem) -> None:
        pass

    def ep_to_coupling(self) -> None:
        pass

    def coupling_to_mechanics(self) -> None:
        pass

    def mechanics_to_coupling(self) -> None:
        pass

    def coupling_to_ep(self) -> None:
        pass

    def setup_assigners(self) -> None:
        from ..datacollector import Assigners

        self._assigners = Assigners()

    def update_prev_mechanics(self) -> None:
        self._t_mechanics_prev = self._time_stepper.t

    def update_prev_ep(self) -> None:
        pass

    def solve_mechanics(self) -> None:
        pass

    def solve_ep(self, interval: typing.Tuple[float, float]) -> None:
        pass

    @property
    def assigners(self) -> datacollector.Assigners:
        return self._assigners

    def save_state(
        self,
        path: typing.Union[str, Path],
        config: typing.Optional[Config] = None,
    ) -> None:
        if config is None:
            logger.warning("Saving state without configuration data")
            config = Config()
        config.coupling_type = self.coupling_type
        io.save_state(
            path=path,
            config=config,
            geo=self.geometry,
            state_params=self.state_params,
        )

    # Strictly speaking this could be a classmethod but we are
    # are not using the class
    @staticmethod
    def from_state(
        path: typing.Union[str, Path],
        drug_factors_file: typing.Union[str, Path] = "",
        popu_factors_file: typing.Union[str, Path] = "",
        disease_state="healthy",
        PCL: float = 1000,
    ):
        return io.load_state(
            path=path,
            drug_factors_file=drug_factors_file,
            popu_factors_file=popu_factors_file,
            disease_state=disease_state,
            PCL=PCL,
        )

    def register_datacollector(self, collector: datacollector.DataCollector) -> None:
        for group_name, group in self.assigners.functions.items():
            for func_name, func in group.items():
                collector.register(group=group_name, name=func_name, f=func)

    def print_mechanics_info(self):
        pass

    def print_ep_info(self):
        pass

    def cell_params(self):
        return {}



from abc import ABC
from abc import abstractmethod
from typing import Dict

from cbcbeat import CardiacCellModel


class BaseCellModel(CardiacCellModel, ABC):
    @staticmethod
    @abstractmethod
    def update_disease_parameters(
        params: Dict[str, float],
        disease_state: str = "healthy",
    ) -> None:
        ...



import contextlib
import warnings
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Optional
from typing import Type
from typing import Union

import cbcbeat
import dolfin
import numpy as np
from dolfin import FiniteElement  # noqa: F401
from dolfin import MixedElement  # noqa: F401
from dolfin import tetrahedron  # noqa: F401
from dolfin import VectorElement  # noqa: F401

from . import geometry
from . import mechanics_model
from . import utils
from .config import Config


logger = utils.getLogger(__name__)


def vs_functions_to_dict(vs, state_names):
    return {
        name: utils.sub_function(vs, index) for index, name in enumerate(state_names)
    }


@contextlib.contextmanager
def h5pyfile(h5name, filemode="r"):
    import h5py
    from mpi4py import MPI

    if h5py.h5.get_config().mpi and dolfin.MPI.size(dolfin.MPI.comm_world) > 1:
        h5file = h5py.File(h5name, filemode, driver="mpio", comm=MPI.COMM_WORLD)
    else:
        if dolfin.MPI.size(dolfin.MPI.comm_world) > 1:
            warnings.warn("h5py is not installed with MPI support")
        h5file = h5py.File(h5name, filemode)
    yield h5file

    h5file.close()


def dict_to_h5(data, h5name, h5group, use_attrs: bool = True):
    with h5pyfile(h5name, "a") as h5file:
        if h5group == "":
            group = h5file
        else:
            group = h5file.create_group(h5group)
        for k, v in data.items():
            if use_attrs:
                group.attrs[k] = v
            else:
                try:
                    group.create_dataset(k, data=v)
                except OSError:
                    logger.warning(
                        f"Unable to save key {k} with data {v} in {h5name}/{h5group}",
                    )


def decode(x):
    if isinstance(x, bytes):
        return x.decode()
    elif isinstance(x, list):
        return [xi for xi in map(decode, x)]
    return x


def h5_to_dict(h5group, use_attrs: bool = True):
    import h5py

    if use_attrs:
        d: Dict[str, Any] = {}
        for k, v in dict(h5group.attrs).items():
            if isinstance(v, np.int64):
                d[k] = int(v)
            elif isinstance(v, np.float64):
                d[k] = float(v)
            else:
                d[k] = v
        return d

    group = {}
    for key, value in h5group.items():
        if isinstance(value, h5py.Dataset):
            v = decode(value[...].tolist())
            group[key] = v

        elif isinstance(value, h5py.Group):
            group[key] = h5_to_dict(h5group[key])

        else:
            raise ValueError(f"Unknown value type {type(value)} for key {key}.")

    return group


def check_file_exists(h5_filename, raise_on_false=True):
    path = Path(h5_filename)
    if not path.is_file():
        if raise_on_false:
            raise FileNotFoundError(f"Path {h5_filename} does not exist")
        return False
    return True


def group_in_file(h5_filename, h5group):
    check_file_exists(h5_filename)
    exists = False

    with h5pyfile(h5_filename) as h5file:
        if h5group in h5file:
            exists = True

    return exists


def mech_heart_to_bnd_cond(mech_heart: mechanics_model.MechanicsProblem):
    if type(mech_heart).__name__ == "RigidMotionProblem":
        return "rigid"
    return "dirichlet"


def serialize_dict(d):
    new_d = {}
    for k, v in d.items():
        if isinstance(v, Path):
            new_d[k] = v.as_posix()
        elif v is None:
            # Skip it
            continue
        elif isinstance(v, dict):
            new_d[k] = serialize_dict(v)
        elif isinstance(v, dolfin.Constant):
            try:
                new_d[k] = float(v)
            except Exception:
                continue
        else:
            new_d[k] = v
    return new_d


def save_state(
    path,
    config: Config,
    geo: geometry.BaseGeometry,
    state_params: Optional[Dict[str, float]] = None,
):
    path = Path(path)
    utils.remove_file(path)

    logger.info(f"Save state to {path}")
    geo.dump(path)
    logger.debug("Save using dolfin.HDF5File")

    logger.debug("Save using h5py")
    dict_to_h5(serialize_dict(config.as_dict()), path, "config")
    if state_params is None:
        state_params = {}
    dict_to_h5(serialize_dict(state_params), path, "state_params")


def load_state(
    path: Union[str, Path],
    drug_factors_file: Union[str, Path] = "",
    popu_factors_file: Union[str, Path] = "",
    disease_state="healthy",
    PCL: float = 1000,
):
    logger.debug(f"Load state from path {path}")
    path = Path(path)
    if not path.is_file():
        raise FileNotFoundError(f"File {path} does not exist")

    logger.debug("Open file with h5py")
    with h5pyfile(path) as h5file:
        config = Config(**h5_to_dict(h5file["config"]))

    if config.coupling_type == "explicit_ORdmm_Land":
        from .models.explicit_ORdmm_Land import EMCoupling
    elif config.coupling_type == "fully_coupled_ORdmm_Land":
        from .models.fully_coupled_ORdmm_Land import EMCoupling  # type: ignore
    elif config.coupling_type == "pureEP_ORdmm_Land":
        from .models.pureEP_ORdmm_Land import EMCoupling  # type: ignore
    elif config.coupling_type == "fully_coupled_Tor_Land":
        from .models.fully_coupled_Tor_Land import EMCoupling  # type: ignore
    else:
        raise ValueError(f"Invalid coupling type: {config.coupling_type}")

    return EMCoupling.from_state(
        path=path,
        drug_factors_file=drug_factors_file,
        popu_factors_file=popu_factors_file,
        disease_state=disease_state,
        PCL=PCL,
    )


def load_initial_conditions_from_h5(
    path: Union[str, Path],
    CellModel: Type[cbcbeat.CardiacCellModel],
):
    path = Path(path)
    if not path.is_file():
        raise FileNotFoundError(f"File {path} does not exist")
    logger.info(f"Loading initial conditions from {path}")

    with h5pyfile(path) as h5file:
        vs_signature = h5file["ep"]["vs"].attrs["signature"].decode()

    mesh = dolfin.Mesh()
    with dolfin.HDF5File(mesh.mpi_comm(), path.as_posix(), "r") as h5file:
        h5file.read(mesh, "/mesh", False)

    VS = dolfin.FunctionSpace(mesh, eval(vs_signature))
    vs = dolfin.Function(VS)

    with dolfin.HDF5File(mesh.mpi_comm(), path.as_posix(), "r") as h5file:
        h5file.read(vs, "/ep/vs")

    return vs_functions_to_dict(
        vs,
        state_names=CellModel.default_initial_conditions().keys(),
    )



import dolfin
import pulse

from . import cell_model
from . import em_model
from .cell_model import ORdmmLandExplicit as CellModel
from .em_model import EMCoupling

loggers = [
    "simcardems.explicit_ORdmm_Land.cell_model.logger",
    "simcardems.explicit_ORdmm_Land.em_model.logger",
]


class ActiveModel(pulse.ActiveModel):
    def __init__(self, coupling: EMCoupling, **kwargs) -> None:
        V = dolfin.FunctionSpace(coupling.geometry.mesh, "CG", 1)
        Ta = dolfin.Function(V)
        super().__init__(
            model="active_stress",
            activation=Ta,
            f0=coupling.geometry.f0,
            s0=coupling.geometry.s0,
            n0=coupling.geometry.n0,
        )


__all__ = [
    "EMCoupling",
    "CellModel",
    "ActiveModel",
    "loggers",
    "em_model",
    "cell_model",
]



from . import cell_model
from . import em_model
from . import explicit_ORdmm_Land
from . import fully_coupled_ORdmm_Land
from . import fully_coupled_Tor_Land
from . import pureEP_ORdmm_Land


def list_coupling_types():
    return [
        "explicit_ORdmm_Land",
        "fully_coupled_ORdmm_Land",
        "fully_coupled_Tor_Land",
        "pureEP_ORdmm_Land",
    ]


loggers = (
    explicit_ORdmm_Land.loggers
    + fully_coupled_ORdmm_Land.loggers
    + pureEP_ORdmm_Land.loggers
    + fully_coupled_Tor_Land.loggers
)


__all__ = [
    "explicit_ORdmm_Land",
    "fully_coupled_ORdmm_Land",
    "fully_coupled_Tor_Land",
    "em_model",
    "cell_model",
    "pureEP_ORdmm_Land",
    "loggers",
]



import json
import typing
from pathlib import Path

from tqdm import tqdm

from . import save_load_functions as io
from . import utils
from .config import Config
from .models import em_model
from .time_stepper import TimeStepper


logger = utils.getLogger(__name__)


class Runner:
    def __init__(
        self,
        config: typing.Optional[Config] = None,
        empty: bool = False,
    ) -> None:
        if config is None:
            config = Config()

        self._config = config
        self.outdir.mkdir(exist_ok=True)
        # Save config to outdir

        def serialize(x):
            if isinstance(x, Path):
                return x.as_posix()
            return x

        (self.outdir / "config.json").write_text(
            json.dumps(self._config.as_dict(), default=serialize),
        )

        from . import set_log_level

        set_log_level(config.loglevel)

        if empty:
            return

        self._reset = not self._config.load_state

        if self._config.load_state and self.state_path.is_file():
            # Load state
            logger.info("Load previously saved state")
            self.coupling = io.load_state(
                path=self.state_path,
                drug_factors_file=self._config.drug_factors_file,
                popu_factors_file=self._config.popu_factors_file,
                disease_state=self._config.disease_state,
                PCL=self._config.PCL,  # Set bcl from cli
            )
        else:
            logger.info("Create a new state")
            # Create a new state
            self.coupling = em_model.setup_EM_model_from_config(self._config)

        self._t0 = self.coupling.t
        self._time_stepper: typing.Optional[TimeStepper] = None
        self._setup_datacollector()
        logger.info(f"Starting at t0={self._t0}")

    @property
    def _dt(self):
        return self._config.dt

    @_dt.setter
    def _dt(self, value):
        self._config.dt = value

    @property
    def state_path(self) -> Path:
        return self.outdir / "state.h5"

    @property
    def outdir(self) -> Path:
        return Path(self._config.outdir)

    @property
    def t(self) -> float:
        if self._time_stepper is None:
            raise RuntimeError("Please create a time stepper before solving")
        return self._time_stepper.t

    @property
    def t0(self) -> float:
        return self._t0

    def _setup_time_stepper(
        self,
        T: float,
        use_ns: bool = True,
        st_progress: typing.Any = None,
    ) -> None:
        self._time_stepper = TimeStepper(
            t0=self._t0,
            T=T,
            dt=self._dt,
            use_ns=use_ns,
            st_progress=st_progress,
        )
        self.coupling.register_time_stepper(self._time_stepper)

    @classmethod
    def from_models(
        cls,
        coupling: em_model.BaseEMCoupling,
        config: typing.Optional[Config] = None,
        reset: bool = True,
    ):
        obj = cls(empty=True, config=config)
        obj.coupling = coupling
        obj._t0 = coupling.t
        obj._reset = reset
        obj._time_stepper = None
        obj._setup_datacollector()
        return obj

    def store(self):
        # Assign u, v and Ca for postprocessing
        self.coupling.assigners.assign()
        self.collector.store(TimeStepper.ns2ms(self.t))

    def _setup_datacollector(self):
        from .datacollector import DataCollector

        self.collector = DataCollector(
            outdir=self.outdir,
            geo=self.coupling.geometry,
            reset_state=self._reset,
            outfilename=self._config.outfilename,
        )
        self.coupling.register_datacollector(self.collector)

    def _solve_mechanics_now(self) -> bool:
        if self._config.mechanics_solve_strategy == "fixed":
            return self.coupling.dt_mechanics > self._config.dt_mech

        self.coupling.assigners.assign_pre()
        norm = self.coupling.assigners.compute_pre_norm()
        return (
            norm >= self._config.mech_threshold
            or self.coupling.dt_mechanics > self._config.dt_mech
        )

    def _post_mechanics_solve(self) -> None:
        # Update previous lmbda
        self.coupling.update_prev_mechanics()
        self.coupling.mechanics_to_coupling()
        self.coupling.coupling_to_ep()

    def _solve_mechanics(self):
        # if self._config.mechanics_use_continuation:
        #     self.mech_heart.solve_for_control(self.coupling.XS_ep)
        # else:
        self.coupling.coupling_to_mechanics()
        self.coupling.solve_mechanics()
        self._post_mechanics_solve()

    def _post_ep(self):
        self.coupling.update_prev_ep()
        self.coupling.ep_to_coupling()

    def save_state(self):
        self.coupling.save_state(path=self.state_path, config=self._config)

    def solve(
        self,
        T: float = Config.T,
        save_freq: int = Config.save_freq,
        show_progress_bar: bool = Config.show_progress_bar,
        st_progress: typing.Any = None,
        default_save_condition: typing.Callable[[int, float, float], bool] = (
            lambda i, T, dt: i > 0 and T >= 40000 and i % int(10000 / dt) == 0
        ),
    ):
        save_it = int(save_freq / self._dt)
        self._setup_time_stepper(T, use_ns=True, st_progress=st_progress)

        pbar = create_progressbar(
            time_stepper=self._time_stepper,
            show_progress_bar=show_progress_bar,
        )

        for i, (t0, t) in enumerate(pbar):
            logger.debug(
                f"Solve EP model at step {i} from {TimeStepper.ns2ms(t0):.2f} ms to {TimeStepper.ns2ms(t):.2f} ms",
            )

            # Solve EP model
            self.coupling.t = TimeStepper.ns2ms(t)
            self.coupling.solve_ep((TimeStepper.ns2ms(t0), TimeStepper.ns2ms(t)))
            self._post_ep()

            if self._solve_mechanics_now():
                logger.debug(f"Solve mechanics model at step {i} from ")
                self._solve_mechanics()

            # Store every 'save_freq' ms
            if i % save_it == 0:
                self.store()

            # Save state every 10 beats if simulation is longer than 40 sec
            if default_save_condition(i, T, self._dt):
                self.coupling.save_state(
                    path=self.outdir.joinpath(f"state_{int(i*self._dt/1000)}beat.h5"),
                    config=self._config,
                )

        self.save_state()


def create_progressbar(
    time_stepper: typing.Optional[TimeStepper] = None,
    show_progress_bar: bool = Config.show_progress_bar,
):
    if time_stepper is None:
        raise ValueError("Please provide a time stepper")
    if show_progress_bar:
        # Show progressbar
        pbar = tqdm(time_stepper, total=time_stepper.total_steps)
    else:
        # Hide progressbar
        pbar = _tqdm(time_stepper, total=time_stepper.total_steps)
    return pbar


class _tqdm:
    def __init__(self, iterable, *args, **kwargs):
        self._iterable = iterable

    def set_postfix(self, msg):
        pass

    def __iter__(self):
        return iter(self._iterable)



from pathlib import Path

import pytest
import simcardems
from simcardems import models


@pytest.mark.slow
def test_explicit_ORdmm_Land(geo, tmpdir):
    coupling = simcardems.models.em_model.setup_EM_model(
        cls_EMCoupling=models.explicit_ORdmm_Land.EMCoupling,
        cls_CellModel=models.explicit_ORdmm_Land.CellModel,
        cls_ActiveModel=models.explicit_ORdmm_Land.ActiveModel,
        geometry=geo,
    )
    config = simcardems.Config(outdir=Path(tmpdir), coupling_type="explicit_ORdmm_Land")
    runner = simcardems.Runner.from_models(coupling=coupling, config=config)
    runner.solve(1.0)
    assert runner.state_path.is_file()
    new_coupling = models.explicit_ORdmm_Land.EMCoupling.from_state(
        path=runner.state_path,
    )
    assert new_coupling == coupling


@pytest.mark.slow
def test_fully_coupled_ORdmm_Land(geo, tmpdir):
    config = simcardems.Config(
        outdir=Path(tmpdir),
        coupling_type="fully_coupled_ORdmm_Land",
    )
    coupling = simcardems.models.em_model.setup_EM_model_from_config(
        config=config,
        geometry=geo,
    )

    runner = simcardems.Runner.from_models(coupling=coupling, config=config)
    runner.solve(1.0)
    assert runner.state_path.is_file()
    new_coupling = models.fully_coupled_ORdmm_Land.EMCoupling.from_state(
        path=runner.state_path,
    )
    assert new_coupling == coupling


@pytest.mark.slow
def test_pure_ep_ORdmm_Land(geo, tmpdir):
    config = simcardems.Config(
        outdir=Path(tmpdir),
        coupling_type="pureEP_ORdmm_Land",
    )
    coupling = simcardems.models.em_model.setup_EM_model_from_config(
        config=config,
        geometry=geo,
    )

    runner = simcardems.Runner.from_models(coupling=coupling, config=config)
    runner.solve(1.0)
    assert runner.state_path.is_file()
    new_coupling = models.pureEP_ORdmm_Land.EMCoupling.from_state(
        path=runner.state_path,
    )
    assert new_coupling == coupling
